# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
DELSIG helpers routines
=======================
"""

import numpy as np
from math import ceil, sqrt
from ._padding import padl, padr
from ._rmsGain import rmsGain
from ..relab import db

__all__ = ["ds_synNTFobj1", "ds_f1f2", "ds_optzeros"]


def ds_synNTFobj1(x, p, osr, f0):
    """
    Objective function for synthesizeNTF.
    """
    z = np.exp(2j*np.pi*(f0+0.5/osr*x))
    if f0 > 0:
        z = padl(z, len(p)//2, np.exp(2j*np.pi*f0))
    z = np.concatenate((z, z.conj()))
    if f0 == 0:
        z = padr(z, len(p), 1)
    f1, f2 = ds_f1f2(osr, f0)

    return db(rmsGain((z, p, 1), f1, f2))


def ds_f1f2(osr=64, f0=0, complex_flag=False):
    """
    Lower and higher extremes of the signal band as normalized frequencies

    Helper function.


    Parameters
    ----------
    osr : float, optional
        the oversamping ratio
    f0 : float, optional
        normalized center frequency for BP modulators, or 0 for LP modulators.
        Defaults to 0.
    complex_flag : bool, optional
        flag indicating if the modulator is quadrature type.

    Returns
    -------
    f1f2 : tuple with two entries corresponding to the lower and
        higher extremes of the signal band.
    """
    if complex_flag:
        f1 = f0-0.5/osr
        f2 = f0+0.5/osr
    else:
        if f0 > 0.25/osr:
            f1 = f0-0.25/osr
            f2 = f0+0.25/osr
        else:
            f1 = 0
            f2 = 0.5/osr
    return f1, f2


def ds_optzeros(n, opt=1):
    """
    Zeros which minimize the in-band noise of a delta-sigma modulator

    Helper function for synthesizeNTF, that returns the zeros as normalized
    angular frequencies


    Parameters
    ----------
    n : int
        the number of optimized zeros to return
    opt : int, optional
        flag for optimized zeros, defaults to 1
        0 -> not optimized
        1 -> optimized
        2 -> optimized with at least one zero at band-center

    Returns
    -------
    zeros : ndarray of reals
        the zeros for the modulator as normalized angular frequencies.

    Notes
    -----
    The zeros are always located on the complex unit circle. As such,
    they are returned as frequencies, not as complex values.

    The zero's frequencies are normalized with respect to the signal
    bandwidth. See also Sec. 4.3.1 in [1]_

    .. [1] Richard Schreier, Gabor C. Temes, "Understanding Delta-Sigma Data
       Converters," IEEE Press and Wiley Interscience, 2005.
    """
    if opt == 0:
        optZeros = np.zeros(ceil(n/2.))
    else:
        if n == 1:
            optZeros = np.asarray([0.])
        elif n == 2:
            if opt == 1:
                optZeros = np.asarray([sqrt(1./3)])
            else:
                optZeros = np.asarray([0.])
        elif n == 3:
            optZeros = np.asarray([sqrt(3./5), 0.])
        elif n == 4:
            if opt == 1:
                discr = sqrt(9./49-3./35)
                tmp = 3./7
                optZeros = np.sqrt([tmp+discr, tmp-discr])
            else:
                optZeros = np.asarray([0., sqrt(5./7)])
        elif n == 5:
            discr = sqrt(25./81-5./21)
            tmp = 5./9
            optZeros = np.sqrt([tmp+discr, tmp-discr, 0.])
        elif n == 6:
            if opt == 1:
                optZeros = np.asarray([0.23862059, 0.66120988, 0.9324696])
            else:
                discr = sqrt(56.)/33
                tmp = 7./11
                optZeros = np.sqrt([0, tmp+discr, tmp-discr])
        elif n == 7:
            optZeros = np.asarray([0, 0.40584371, 0.74153078, 0.94910785])
        elif n == 8:
            if opt == 1:
                optZeros = np.asarray([0.18343709, 0.52553345, 0.79666684,
                                       0.96028993])
            else:
                optZeros = np.asarray([0, 0.50563161, 0.79017286, 0.95914731])
        elif n == 9:
            optZeros = np.asarray([0, 0.32425101, 0.61337056, 0.83603082,
                                   0.9681602])
        elif n == 10:
            if opt == 1:
                optZeros = np.asarray([0.1834370913, 0.5255334458,
                                       0.7966668433, 0.9602899327])
            else:
                optZeros = np.asarray([0, 0.41572267, 0.67208682, 0.86238894,
                                       0.97342121])
        elif n == 11:
            optZeros = np.asarray([0, 0.26953955, 0.51909468, 0.73015137,
                                   0.88706238, 0.97822864])
        elif n == 12:
            if opt == 1:
                optZeros = np.asarray([0.12523875, 0.36783403, 0.58731921,
                                       0.7699033, 0.90411753, 0.9815607])
            else:
                optZeros = np.asarray([0, 0.35222363, 0.58006251, 0.76647993,
                                       0.90281326, 0.98132047])
        elif n == 13:
            optZeros = np.asarray([0, 0.23045331, 0.44849063, 0.64234828,
                                   0.8015776, 0.91759824, 0.98418306])
        elif n == 14:
            if opt == 1:
                optZeros = np.asarray([0.10806212, 0.31911586, 0.51525046,
                                       0.68729392, 0.82720185, 0.92843513,
                                       0.98628389])
            else:
                optZeros = np.asarray([0, 0.30524384, 0.50836649, 0.6836066,
                                       0.82537239, 0.92772336, 0.98615167])
        else:
            raise ValueError('Optimized zeros for n>14 are not available.')

    # Sort the zeros and replicate them.
    z = np.sort(optZeros)
    optZeros = np.zeros(n)
    m = 0
    if (n % 2) == 1:
        optZeros[0] = z[0]
        z = z[1:]
        m = m+1
    for i in range(len(z)):
        optZeros[m] = z[i]
        optZeros[m+1] = -z[i]
        m = m+2

    return optZeros



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
import cvxopt as opt
from ..defs import *
from ..scalars import cvxpy_obj
from ..scalars import cvxpy_scalar_var
from ..utils import vstack

# Class definition
class cvxpy_kl_div_epi(object):
    r"""
    | :math:`\{(x,y,t) \in \mathbb{R}_{++}^n \\times \mathbb{R}_{++}^n
      \\times \mathbb{R} \ |
      \ \sum_{i=1}^n (x_i \mbox{log} (x_i/y_i) - x_i + y_i) \leq t \}`.
    """

    # Method: __init__
    def __init__(self):
        """
        Class constructor.
        """

        self.type = SET
        self.name = 'kl_div_epi'
        self.expansion_type = DIF

    # Method: valid_shape
    def valid_shape(self,shape):
        m,n = shape
        return n == 1 and m >= 3 and m%2 == 1

    # Method: __str__
    def __str__(self):

        return self.name

    # Method: _construct
    def _construct(self,el,mp,p):

        n = int((el.shape[0]-1.)/2.)
        u = vstack([el[0:n,0]])
        v = vstack([el[n:2*n,0]])
        t = el[2*n,0]

        # Prepare u and v
        for i in range(0,n):
            if type(u[i,0]) is cvxpy_obj:
                u[i,0] = u[i,0].value
            if type(v[i,0]) is cvxpy_obj:
                v[i,0] = v[i,0].value

        # Prepare t
        if type(t) is cvxpy_obj:
            t = t.value

        # f
        def f(x):
            temp = 0
            for i in range(0,n):
                if type(u[i,0]) is cvxpy_scalar_var:
                    u_i = x[mp[u[i,0]]]
                else:
                    u_i = u[i,0]
                if type(v[i,0]) is cvxpy_scalar_var:
                    v_i = x[mp[v[i,0]]]
                else:
                    v_i = v[i,0]
                temp += u_i*np.log(u_i*1./v_i)-u_i+v_i
            if type(t) is cvxpy_scalar_var:
                return temp - x[mp[t]]
            else:
                return temp - t

        # grad f
        def grad_f(x):
            g = opt.spmatrix(0.0,[],[],(p,1))
            for i in range(0,n):
                if type(u[i,0]) is cvxpy_scalar_var:
                    u_i = x[mp[u[i,0]]]
                else:
                    u_i = u[i,0]
                if type(v[i,0]) is cvxpy_scalar_var:
                    v_i = x[mp[v[i,0]]]
                else:
                    v_i = v[i,0]
                if type(u[i,0]) is cvxpy_scalar_var:
                    g[mp[u[i,0]]] = np.log(u_i*1./v_i)
                if type(v[i,0]) is cvxpy_scalar_var:
                    g[mp[v[i,0]]] = (-u_i*1./v_i)+1.
            if type(t) is cvxpy_scalar_var:
                g[mp[t]] = -1.
            return g

        # hess f
        def hess_f(x):
            h = opt.spmatrix(0.0,[],[],(p,p))
            for i in range(0,n):
                if type(u[i,0]) is cvxpy_scalar_var:
                    u_i = x[mp[u[i,0]]]
                else:
                    u_i = u[i,0]
                if type(v[i,0]) is cvxpy_scalar_var:
                    v_i = x[mp[v[i,0]]]
                else:
                    v_i = v[i,0]
                if type(u[i,0]) is cvxpy_scalar_var:
                    h[mp[u[i,0]],mp[u[i,0]]] = 1./u_i
                if type(v[i,0]) is cvxpy_scalar_var:
                    h[mp[v[i,0]],mp[v[i,0]]] = u_i*1./(v_i**2.)
                if (type(u[i,0]) is cvxpy_scalar_var and
                    type(v[i,0]) is cvxpy_scalar_var):
                    w = -1./v_i
                    h[mp[u[i,0]],mp[v[i,0]]] = w
                    h[mp[v[i,0]],mp[u[i,0]]] = w
            return h

        # ind f
        def ind_f(x):
            for i in range(0,n):
                if type(u[i,0]) is cvxpy_scalar_var:
                    if x[mp[u[i,0]]] <= 0:
                        return False
                else:
                    if u[i,0] <= 0:
                        return False
                if type(v[i,0]) is cvxpy_scalar_var:
                    if x[mp[v[i,0]]] <= 0:
                        return False
                else:
                    if v[i,0] <= 0:
                        return False
            return True

        # Return functions
        return f, grad_f, hess_f, ind_f

# Create instance
kl_div_epi = cvxpy_kl_div_epi()



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.


# Following part is deprecated

from __future__ import division, print_function

from ..merit_factors import quantization_noise_gain
from warnings import warn
from ...exceptions import PyDsmDeprecationWarning

__all__ = ["quantization_weighted_noise_gain"]


def quantization_weighted_noise_gain(NTF, w=None, bounds=(0, 0.5),
                                     **options):
    """
    Alias of :func:`pydsm.NTFdesign.quantization_noise_gain`.

    .. deprecated:: 0.11.0
        Function has been moved to the :mod:`pydsm.NTFdesign` module.
    """
    warn("Function superseded by quantization_noise_gain in "
         "NTFdesign module", PyDsmDeprecationWarning)
    return quantization_noise_gain(NTF, w, bounds, **options)



# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

# Helper inline functions for cython simulateDSM code

cdef inline double dbl_sat(double x, double a, double b):
    return a if x <= a else b if x>=b else x

cdef inline void ds_quantize(int N, double* y, int y_stride, \
    int* n, int n_stride, \
    double* v, int v_stride):
    """Quantize a signal according to a given number of levels."""
    cdef int qi
    cdef double L
    for qi in range(N):
        if n[qi*n_stride] % 2 == 0:
            v[qi*v_stride] = 2*floor(0.5*y[qi*y_stride])+1
        else:
            v[qi*v_stride] = 2*floor(0.5*(y[qi*y_stride]+1))
        L = n[qi*n_stride]-1
        v[qi*v_stride]=dbl_sat(v[qi*v_stride],-L,L)

cdef inline void track_vabsmax(int N,\
    double* vabsmax, int vabsmax_stride,\
    double* x, int x_stride):
    cdef int i
    cdef double absx
    for i in range(N):
        absx=fabs(x[i*x_stride])
        if absx > vabsmax[i*vabsmax_stride]:
            vabsmax[i*vabsmax_stride]=absx

cdef inline double *dbldata(np.ndarray arr):
    return <double *>np.PyArray_DATA(arr)

cdef inline int *intdata(np.ndarray arr):
    return <int *>np.PyArray_DATA(arr)


#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..sets import *
from ..utils import *
from ..interface import *
from ..arrays import cvxpy_matrix

# Lambda max
def lambda_max(x):
    r"""
    | :math:`\mbox{lambda\_max} :
      \mathbb{S}^n \\to \mathbb{R},
      \ \mbox{lambda\_max}(X) = \mbox{sup}\{ y^TXy \ | \ \|y\|_2 = 1 \}`.
    | Convex.

    :param x: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: number or
             :ref:`tree<tree_obj>`.
    """

    # Check type
    if (np.isscalar(x) or
        type(x).__name__ in SCALAR_OBJS):
        return x
    elif (type(x) is not cvxpy_matrix and
          type(x).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid argument type')

    # Check shape
    (m,n) = x.shape
    if m!=n:
        raise ValueError('Invalid argument dimensions')

    # Construct objective and constraints
    z = variable(m,m)
    t = variable()
    obj = t
    constr = [belongs(t*eye(n)-z,semidefinite_cone)]

    # Add symmetry constraint if input is not numeric
    if type(x).__name__ in ARRAY_OBJS:
        constr += [equals(z,z.T)]

    # Check symmetry if input is numeric
    elif not np.allclose(x,x.T):
        return np.inf

    # Construct program
    p = program(minimize(obj),
                constr,
                [z],
                name='lambda_max')

    # Return program
    return p(x)



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Collection of DELSIG style decibel routines
===========================================
"""

from __future__ import division

import numpy as np

__all__ = ["dbv", "dbp", "dbm", "undbv", "undbp", "undbm"]


def dbv(x):
    """
    Converts a voltage (or current) ratio to dB.

    Parameters
    ----------
    x : real or array_like of reals
        argument should be strictly positive

    Returns
    -------
    y : real or array_like of reals
        20*log10(x)
    """
    x = np.asarray(np.abs(x))
    y = np.asarray(-np.inf*np.ones_like(x))
    nonzero = np.asarray(x != 0)
    y[nonzero] = 20*np.log10(x[nonzero])
    return y[()]


def dbp(x):
    """
    Converts a power ratio to dB.

    Parameters
    ----------
    x : real or array_like of reals
        argument should be strictly positive

    Returns
    -------
    y : real or array_like of reals
        10*log10(x)
    """
    x = np.asarray(np.abs(x))
    y = np.asarray(-np.inf*np.ones_like(x))
    nonzero = np.asarray(x != 0)
    y[nonzero] = 10*np.log10(x[nonzero])
    return y[()]


def dbm(v, R=50.):
    """
    Converts argument from rms voltage to dBm.

    Parameters
    ----------
    v : real or array_like of reals
        rms value to be converted in dBm
    R : real, optional
        value of test resistor (defaults to 50)

    Returns
    -------
    y : real or array_like of reals
        dBm value corresponding to the power provided by the rms voltag v
        on the test resistor R, normalized over the reference value of 1mW
    """
    v = np.asarray(np.abs(v))
    y = np.asarray(-np.inf*np.ones_like(v))
    nonzero = np.asarray(v != 0)
    y[nonzero] = 10*np.log10(v[nonzero]**2/R)+30
    return y[()]


def undbv(x):
    """
    Converts argument from db to a voltage ratio.

    Inverse of `dbv`.

    Parameters
    ----------
    x : real or array_like of reals
        value to convert

    Returns
    -------
    y : real or array_like of reals
        10^(x/20)
    """
    return 10**(np.asarray(x)/20.)


def undbp(x):
    """
    Converts argument from db to a power ratio.

    Inverse of `dbp`.

    Parameters
    ----------
    x : real or array_like of reals
        value to convert

    Returns
    -------
    y : real or array_like of reals
        10^(x/10)
    """
    return 10**(np.asarray(x)/10.)


def undbm(x, R=50.):
    """
    Converts argument from dBm power to rms voltage.

    Inverse of `dbm`.

    Parameters
    ----------
    x : real or array_like of reals
        value to convert
    R : real, optional
        test resistor

    Returns
    -------
    y : real or array_like of reals
        rms voltage that applied over a resistor R gives a power equal to x
        in dBm (i.e. referred to 1mW)
    """
    return np.sqrt(R*10**((x-30)/10.))



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np

__all__ = ['partitionABCD']


def partitionABCD(ABCD, m=None):
    u"""Partition ABCD ss description in A, B, C, D for an m-input sytem

    Parameters
    ----------
    ABCD : matrix like 2D
        input ABCD matrix providing the state space representation
    m : int or None, optional
        number of inputs of state space system. If set to None, a guess
        is made from the ABCD matrix size (defaults to None).

    Returns
    -------
    (A, B, C, D) : tuple of 2D ndarrays
        the individual matrices composing ABCD

    Notes
    -----
    When the number of inputs is automatically guessed, the minimum
    acceptable value is take, considering the the system must have at least
    an input and an output.

    If ABCD is an ndarray, then its content is not copied in A, B, C, D.
    """
    ABCD = np.asarray(ABCD)
    if m is None:
        n = min(ABCD.shape)-1
    else:
        n = ABCD.shape[1]-m
    return (ABCD[:n, :n], ABCD[:n, n:], ABCD[n:, :n], ABCD[n:, n:])



# -*- coding: utf-8 -*-

# Copyright (c) 2013, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
CLANS NTF design method
=======================

CLANS is the Closed-Loop Analysis of Noise Shaping Coders method
proposed by J. G. Kenney and L. R. Carley in
'Design of Multibit Noise Shaping Data Converters'
Analog Integrated Circuits and Signal Processing Vol. 3 pp 259-272 (1993)
"""

import numpy as np
from scipy.optimize import minimize
from ._synthesizeNTF import synthesizeNTF
from ..relab import cplxpair
from ._tf import evalTF
from ..ir import impulse_response
from ._dsclansNTF import dsclansNTF
from ..utilities import digested_options

__all__ = ["clans"]


def clans(order=4, osr=64, nq=5, rmax=0.95, opt=0, **options):
    u"""Synthesize the NTF for a ΔΣM w/ multibit quantizer by the CLANS method.

    This function is based on the CLANS method (Closed-Loop Analysis of Noise
    Shaping Coders) by J. G. Kenney and L. R. Carley, described in
    'Design of Multibit Noise Shaping Data Converters'
    Analog Integrated Circuits and Signal Processing Vol. 3 pp 259-272 (1993).

    The method is only suitable for low pass modulators.

    Parameters
    ----------
    order : int, optional
        the order of the modulator, defaults to 4
    osr : float, optional
        the oversamping ratio (based on the actual signal bandwidth).
        Defaults to 64
    nq : int, optional
        the number of quantization levels, defaults to 5
    rmax : real, optional
        the maximum radius of the NTF poles
    opt : int, optional
        opt parameter passed to ``synthesizeNTF`` which is used to compute the
        NTF zero location and the initial pole location for the optimizer

    Other Parameters
    ----------------
    show_progress : bool, optional
        provide extended output, default is False and can be updated by
        changing the function ``default_options`` attribute.
    slsqp_opts : dictionary, optional
        Parameters passed to the ``fmin_slsqp`` optimizer. Allowed options are:

        ``maxiters``
            Maximum number of iterations (defaults to 100)
        ``ftol``
            Precision goal for the value of f in the stopping criterion
            (defaults to 1e-6)
        ``eps``
            Step size used for numerical approximation of the jacobian
            (defaults to 1.4901161193847656e-08)

        Do not use other options since they could break the minimizer in
        unexpected ways. Defaults can
        be set by changing the function ``default_options`` attribute.

    Returns
    -------
    ntf : tuple
        noise transfer function in zpk form.

    Raises
    ------
    ValueError
        'Error. f0 must be less than 0.5' if f0 is out of range

        'Order must be even for a bandpass modulator.' if the order is
        incompatible with the modulator type.

        'The opt vector must be of length xxx' if opt is used to explicitly
        pass the NTF zeros and these are in the wrong number.

    Notes
    -----
    This is actually an adaptation of the CLANS method. In the
    original paper about CLANS, NTF zeros are placed at z=1 (with the
    indication that other placements are possible). Here, the zeros are
    placed where ``synthesizeNTF`` places them.

    The computation is based on a nonlinear, nonlinearly constrained
    optimization. Since the optimizer used here is different from the
    optimizer used in other toolboxes implementing this function,
    the results may differ. The current optimizer is ``SLSQP``.

    The function internally calls ``synthesizeNTF``, and rises the same
    exceptions as ``synthesizeNTF``.
    """
    # Manage optional parameters
    opts = digested_options(options, clans.default_options,
                            ['show_progress'], ['slsqp_opts'])
    slsqp_opts = opts['slsqp_opts']
    slsqp_opts['disp'] = opts.get('show_progress', False)
    # Create the initial guess
    NTF = synthesizeNTF(order, osr, opt, 1+nq, 0)
    Hz = NTF[0]
    # x is the vector to optimize, accounting for the poles of the NTF
    # Rather than putting in x the poles of the NTF, the NTF is converted
    # into a CT form via a bilinear transform, then the frequency and damping
    # factor of each pole couple is recorded. This allows the constraint
    # that complex poles must be conjugated in couples to be automatically
    # managed. If the number of poles is odd, then there is necessarily a pole
    # that remains single, which should be treated differently, by storing
    # its frequency only.
    # Some notes...
    # 1. When practicing the bilinear transform to pass into the s-domain
    #    rather than using the poles of the NTF, one uses a version of such
    #    poles scaled by 1/rmax
    # 2. Rather than storing the poles frequency and damping factor, their
    #    square root is saved
    x = np.zeros(order)
    odd = order % 2
    poles = cplxpair(NTF[1])
    poles = poles[::-1]
    if odd:
        # Treat first pole (that is necessarily real)
        z = poles[0].real/rmax
        if np.abs(z) > 1:
            # Project poles outside rmax onto the circle with radius rmax
            # NB. Test has been fixed wrt DELSIG, where it always fails.
            z = z/np.abs(z)
        # Bilinear transform (modulus a mult. coeff)
        s = (z-1)/(z+1)
        x[0] = np.sqrt(-s)
    # All other poles are assumed to be complex.
    # N.B. This may be a wrong assumption, though. There may be more real
    # poles.
    for i in range(odd, order, 2):
        z = poles[i:i+2]/rmax
        if np.any(np.abs(z) > 1):
            # Project poles outside rmax onto the circle with radius rmax
            z = z/np.abs(z)
        # Bilinear transform (modulus a mult. coeff)
        s = (z-1)/(z+1)
        coeffs = np.poly(s).real
        wn = np.sqrt(coeffs[2])
        zeta = coeffs[1]/(2*wn)
        x[i] = np.sqrt(zeta)
        x[i+1] = np.sqrt(wn)

    # Run the optimizer
    x = minimize(_dsclansObj6a, x, args=(order, osr, nq, rmax, Hz),
                 method='SLSQP',
                 constraints={'type': 'ineq',
                              'fun': _dsclansObj6b,
                              'args': (order, osr, nq, rmax, Hz)},
                 options=slsqp_opts)['x']
    return dsclansNTF(x, order, rmax, Hz)

clans.default_options = {'show_progress': False,
                         'slsqp_opts': {'maxiter': 100,
                                        'ftol': 1e-06,
                                        'eps': 1.4901161193847656e-08}}


def _dsclansObj6a(x, order, osr, nq, rmax, Hz):
    # This is the objective function for clans
    # it returns the magnitude of H at the band edge.
    H = dsclansNTF(x, order, rmax, Hz)
    f = np.abs(evalTF(H, np.exp(1j*np.pi/osr)))
    return f


def _dsclansObj6b(x, order, osr, nq, rmax, Hz):
    # This is the constraint function for clans
    # it returns ||h||_1 - nq
    H = dsclansNTF(x, order, rmax, Hz)
    # NB. In the following use dB=60 rather than m=100, maybe
    g = np.sum(np.abs(impulse_response(H, m=100)))-1-nq
    # With our optimizer, this needs to be inverted
    return -g



# -*- coding: utf-8 -*-


# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

u"""
Computation of NTF related merit factors (:mod:`pydsm.NTFdesign.merit_factors`)
===============================================================================

This module provides functions for the computation of NTF merit factors

.. currentmodule:: pydsm.NTFdesign.merit_factors


Functions
---------

.. autosummary::
   :toctree: generated/

   quantization_noise_gain  -- NTF quantization noise power gain
"""

from __future__ import division, print_function

import numpy as np
from scipy.integrate import quad
from ..delsig import evalTF
from ..utilities import digested_options

__all__ = ["quantization_noise_gain"]


def quantization_noise_gain(NTF, w=None, bounds=(0, 0.5), avg=False,
                            **options):
    r"""
    Compute the NTF quantization noise power gain.

    Parameters
    ----------
    NTF : tuple
        NTF definition in zpk or nd form
    w : callable with argument f in [0,1/2] or None or tuple
            * if function: noise weighting function
            * if None: no weighting is applied
            * if filter definition as zpk or ba tuple: weighting is implicitly
              provided by the filter
    bounds : 2 elements tuple, optional
        the frequency range where the noise gain is computed. Defaults to
        (0, 0.5)
    avg: bool, optional
        If True, rather than returning the overall noise gain, the function
        returns the average noise gain over the bandwidth.

    Returns
    -------
    a : real
        noise power gain

    Other parameters
    ----------------
    quad_opts : dictionary, optional
        Parameters to be passed to the ``quad`` function used internally as
        an integrator. Allowed options are ``epsabs``, ``epsrel``, ``limit``,
        ``points``. Do not use other options since they could break the
        integrator in unexpected ways. Defaults can be set by changing the
        function ``default_options`` attribute.

    Notes
    -----
    The computation is practiced as

    .. math::
        2\int_{f=0}^{\frac{1}{2}}
        \left|\mathit{NTF}
        \left(\mathrm{e}^{\mathrm{i} 2\pi f}\right)\right|^2 w(f) \,
        \mathrm{d}f

    Use an on-off weighting function :math:`w(f)` for multiband evaluation.

    In case the weighting function has discontinuities, report them to the
    integrator via the ``quad_points`` parameter.

    See Also
    --------
    scipy.integrate.quad : for the meaning of the integrator parameters.
    """
    # Manage parameters
    if w is None:
        w = lambda f: 1.
    elif type(w) is tuple and 2 <= len(w) <= 3:
        h = w
        w = lambda f: np.abs(evalTF(h, np.exp(2j*np.pi*f)))**2
    # Manage optional parameters
    opts = digested_options(options, quantization_noise_gain.default_options,
                            [], ['quad_opts'])
    # Compute
    c = 1/(bounds[1]-bounds[0]) if avg else 2.
    return c*quad(lambda f: np.abs(evalTF(NTF, np.exp(2j*np.pi*f)))**2*w(f),
                  bounds[0], bounds[1], **opts["quad_opts"])[0]

quantization_noise_gain.default_options = {"quad_opts": {"epsabs": 1E-14,
                                                         "epsrel": 1E-9,
                                                         "limit": 100,
                                                         "points": None}}



# -*- coding: utf-8 -*-

# Copyright (c) 2013, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np

__all__ = ['dsclansNTF']


def dsclansNTF(x, order, rmax, Hz):
    """
    Convert CLANS parameters into an NTF in ZPK form
    """
    # Helper function to convert clans parameters into an NTF in zpk form
    # i.e. translate x into H.
    #
    # Wrt the original code proposal in the article by Kenney and Carley
    # the relationship between (zeta, wn) and x has been changed (as in the
    # code by R. Schreier), in order to guarantee left-half-plane roots for
    # the s polynomial
    Hp = np.zeros(len(Hz), dtype=complex)
    odd = order % 2
    if odd:
        s = -x[0]**2
        # Bilinear transform
        Hp[0] = rmax*(1+s)/(1-s)
    for i in range(odd, order, 2):
        zeta = x[i]**2
        wn = x[i+1]**2
        s = np.roots((1, 2*zeta*wn, wn**2))
        # Bilinear transform
        Hp[i:i+2] = rmax*(1+s)/(1-s)
    return (Hz, Hp, 1)



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
from pydsm.delsig import padl, padr, padt, padb

__all__ = ["TestSynthesizePAD"]


class TestSynthesizePAD:

    def setUp(self):
        pass

    def test_padl_vector(self):
        v = np.asarray([1, 2, 3])
        ve = np.asarray([4, 4, 1, 2, 3])
        np.testing.assert_equal(padl(v, 5, 4), ve)

    def test_padr_vector(self):
        v = np.asarray([1, 2, 3])
        ve = np.asarray([1, 2, 3, 0, 0])
        np.testing.assert_equal(padr(v, 5), ve)

    def test_padl_array2d(self):
        v = np.asarray([[1, 2, 3], [4, 5, 6]])
        ve = np.asarray([[4, 4, 1, 2, 3], [4, 4, 4, 5, 6]])
        np.testing.assert_equal(padl(v, 5, 4), ve)

    def test_padr_array2d(self):
        v = np.asarray([[1, 2, 3], [4, 5, 6]])
        ve = np.asarray([[1, 2, 3, 0, 0], [4, 5, 6, 0, 0]])
        np.testing.assert_equal(padr(v, 5), ve)

    def test_padt_array2d(self):
        v = np.asarray([[1, 2, 3], [4, 5, 6]])
        ve = np.asarray([[4, 4, 4], [4, 4, 4], [1, 2, 3], [4, 5, 6]])
        np.testing.assert_equal(padt(v, 4, 4), ve)

    def test_padb_array2d(self):
        v = np.asarray([[1, 2, 3], [4, 5, 6]])
        ve = np.asarray([[1, 2, 3], [4, 5, 6], [0, 0, 0], [0, 0, 0]])
        np.testing.assert_equal(padb(v, 4), ve)

if __name__ == '__main__':
    run_module_suite()



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

"""
Fast simulator for a generic delta sigma modulator using external cblas
=======================================================================
"""

import numpy as np
cimport numpy as np
import scipy as sp
__import__('scipy.signal')
from libc.math cimport floor, fabs

cdef extern from "cblas.h":
    enum CBLAS_ORDER:     CblasRowMajor, CblasColMajor
    enum CBLAS_TRANSPOSE: CblasNoTrans, CblasTrans, CblasConjTrans
    void cblas_dgemv(CBLAS_ORDER order, \
        CBLAS_TRANSPOSE TransA, int M, int N,\
        double alpha, double *A, int lda,\
        double *X, int incX,\
        double beta, double *Y, int incY)
    void cblas_dcopy(int N, double *X, int incX,\
        double *Y, int incY)

include '_simulateDSM_helper.pxi'

def simulateDSM(np.ndarray u, arg2, nlev=2, x0=0,
                int store_xn=False, int store_xmax=False, int store_y=False):

    # Make sure that nlev is a 1D int array
    cdef np.ndarray c_nlev
    try:
        c_nlev = np.asarray(nlev, dtype=np.int32)
        if c_nlev.ndim > 1:
            raise TypeError()
        c_nlev=c_nlev.reshape(1)
    except (ValueError, TypeError):
         raise ValueError(\
            "Invalid argument: nlev must be convertible into a 1D int array")

    # Make sure that input is a matrix
    cdef np.ndarray c_u
    try:
        c_u = np.asarray(u, dtype=np.float64, order='C')
        if c_u.ndim > 2:
            raise TypeError()
        if c_u.ndim == 1:
            c_u = c_u.reshape(1, -1)
    except (ValueError, TypeError):
        raise ValueError(\
            "Invalid argument: u must be convertible into a 2D float array")

    cdef int nu = c_u.shape[0]
    cdef int nq = c_nlev.shape[0]

    cdef int order

    try:
        if type(arg2)==tuple and len(arg2)==3:
            # Assume ntf in zpk form
            ntf_z=np.asarray(arg2[0], dtype=np.complex128)
            ntf_p=np.asarray(arg2[1], dtype=np.complex128)
            ntf_k=float(arg2[2])
            if ntf_z.ndim !=1 or ntf_p.ndim != 1:
                raise TypeError()
            form = 2
            order = ntf_z.shape[0]
        else:
            # Assume ABCD form
            ABCD = np.asarray(arg2, dtype=np.float64)
            if ABCD.ndim!=2:
                raise TypeError()
            if ABCD.shape[1] != nu+ABCD.shape[0]:
                raise TypeError()
            form = 1
            order = ABCD.shape[0]-nq
    except (ValueError, TypeError):
        raise ValueError('Incorrect modulator specification')

    cdef np.ndarray c_x0, c_x0_temp
    # Assure that the state is a column vector
    try:
        if np.isscalar(x0) and x0 == 0:
            c_x0 = np.zeros((order, 1), dtype=np.float64)
        else:
            c_x0 = np.array(x0, dtype=np.float64, order='C')
            if c_x0.ndim < 1 or c_x0.ndim > 2:
                raise TypeError()
            c_x0=c_x0.reshape(-1, 1)
            if c_x0.shape[0]!=order:
                raise TypeError()
    except (ValueError, TypeError):
        raise ValueError('Incorrect initial condition specification')
    c_x0_temp = np.empty_like(c_x0)

    cdef np.ndarray A, B1, B2, C, D1
    # Build ISO Model
    # note that B=hstack((B1, B2))
    if form == 1:
        A = np.asarray(ABCD[0:order, 0:order], dtype=np.float64, order='C')
        B1 = np.asarray(ABCD[0:order, order:order+nu],\
            dtype=np.float64, order='C')
        B2 = np.asarray(ABCD[0:order, order+nu:order+nu+nq],\
            dtype=np.float64, order='C')
        C = np.asarray(ABCD[order:order+nq, 0:order],\
            dtype=np.float64, order='C')
        D1 = np.asarray(ABCD[order:order+nq, order:order+nu], \
            dtype=np.float64, order='C')
    else:
        # Seek a realization of -1/H
        A, B2, C, D2 = sp.signal.zpk2ss(ntf_p, ntf_z, -1)
        C=C.real
        # Transform the realization so that C = [1 0 0 ...]
        Sinv = sp.linalg.orth(np.hstack((np.transpose(C), np.eye(order))))/ \
            np.linalg.norm(C)
        S = sp.linalg.inv(Sinv)
        C = np.dot(C, Sinv)
        if C[0, 0] < 0:
            S = -S
            Sinv = -Sinv
        A = np.asarray(S.dot(A).dot(Sinv), dtype=np.float64, order='C')
        B2 = np.asarray(np.dot(S, B2), dtype=np.float64, order='C')
        C = np.asarray(np.hstack(([[1.]], np.zeros((1,order-1)))),\
            dtype=np.float64, order='C')
        # C=C*Sinv;
        # D2 = 0;
        # !!!! Assume stf=1
        B1 = -B2
        D1 = np.asarray(1., dtype=np.float64)
        #B = np.hstack((B1, B2))

    # N is number of input samples to deal with
    cdef int N = c_u.shape[1]
    # v is output vector
    cdef np.ndarray v = np.empty((nq, N), dtype=np.float64)
    cdef np.ndarray y
    if store_y:
        # Need to store the quantizer input
        y = np.empty((nq, N), dtype=np.float64)
    else:
        y = np.empty((0,0), dtype=np.float64)
    cdef np.ndarray xn
    if store_xn:
        # Need to store the state information
        xn = np.empty((order, N), dtype=np.float64)
    cdef np.ndarray xmax
    if store_xmax:
        # Need to keep track of the state maxima
        xmax = np.abs(c_x0)
    else:
        xmax = np.empty(0, dtype=np.float64)

    # y0 is output before the quantizer
    cdef np.ndarray y0 = np.empty(nq, dtype=np.float64)

    cdef int i
    for i in xrange(N):
        # Compute y0 = np.dot(C, c_x0) + np.dot(D1, u[:, i])
        cblas_dgemv(CblasRowMajor, CblasNoTrans, nq, order,\
            1.0, dbldata(C), order, \
            dbldata(c_x0), 1, \
            0.0, dbldata(y0), 1)
        cblas_dgemv(CblasRowMajor, CblasNoTrans, nq, nu,\
            1.0, dbldata(D1), nu, \
            dbldata(c_u)+i, N, \
            1.0, dbldata(y0), 1)
        if store_y:
            #y[:, i] = y0[:]
            cblas_dcopy(nq, dbldata(y0), 1,\
            dbldata(y)+i, N)
        ds_quantize(nq, dbldata(y0), 1, \
            intdata(c_nlev), 1, \
            dbldata(v)+i, N)
        # Compute c_x0 = np.dot(A, c_x0) +
        #   np.dot(B, np.vstack((u[:, i], v[:, i])))
        cblas_dgemv(CblasRowMajor, CblasNoTrans, order, order,\
            1.0, dbldata(A), order, \
            dbldata(c_x0), 1,\
            0.0, dbldata(c_x0_temp), 1)
        cblas_dgemv(CblasRowMajor, CblasNoTrans, order, nu,\
            1.0, dbldata(B1), nu, \
            dbldata(c_u)+i, N, \
            1.0, dbldata(c_x0_temp), 1)
        cblas_dgemv(CblasRowMajor, CblasNoTrans, order, nq,\
            1.0, dbldata(B2), nq, \
            dbldata(v)+i, N, \
            1.0, dbldata(c_x0_temp), 1)
        # c_x0[:,1] = c_x0_temp[:,1]
        cblas_dcopy(order, dbldata(c_x0_temp), 1,\
            dbldata(c_x0), 1)
        if store_xn:
            # Save the next state
            #xn[:, i] = c_x0
            cblas_dcopy(order, dbldata(c_x0), 1,\
            dbldata(xn)+i, N)
        if store_xmax:
            # Keep track of the state maxima
            # xmax = np.max((np.abs(x0), xmax), 0)
            track_vabsmax(order, dbldata(xmax), 1,\
                dbldata(c_x0), 1)
    if not store_xn:
        xn = c_x0
    return v.squeeze(), xn.squeeze(), xmax, y.squeeze()



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from .defs import *
from .arrays import *
from .scalars import *
from .constraints import *
from .procedures.expand import expand
from .procedures.re_eval import re_eval
from .procedures.solve_prog import solve_prog

#***********************************************************************#
# Class definition: cvxpy_program                                       #
#***********************************************************************#
class cvxpy_program(object):

    # Method: __init__
    def __init__(self,action,objective,constraints,
                formals,options,name):
        """
        Class constructor.

        :param action: Keyword (see cvxpy.def).
        :param objective: Scalar expression.
        :param constraints: List of constraints.
        :param formals: List of variables or parameters.
        :para options: Dictionary.
        :param name: String.
        """

        # Attributes
        self.name = name
        self.type = FUNCTION
        self.action = action
        self.constraints = cvxpy_list(constraints)

        # Objective
        if np.isscalar(objective):
            self.objective = cvxpy_obj(CONSTANT,objective,str(objective))
        elif (type(objective) is cvxpy_obj or
              type(objective).__name__ in SCALAR_OBJS):
            self.objective = objective
        else:
            raise TypeError('Invalid objective')

        # Formals
        for x in formals:
            if type(x).__name__ not in FORMAL_OBJS:
                raise TypeError('Invalid formal')
        self.formals = cvxpy_list(formals)

        # Options
        if options != None:
            self.options = options
        else:
            self.options = CONFIGURATION.copy()

    # Method: _get_expanded_objects
    def _get_expanded_objects(self,args):
        """
        Expands list of objects to list of scalar objects.

        :param args: List of objects.
        """

        args_expanded = []
        for x in args:
            if (np.isscalar(x) or
                type(x).__name__ in SCALAR_OBJS):
                args_expanded += [x]
            elif (type(x) is cvxpy_matrix or
                  type(x).__name__ in ARRAY_OBJS):
                for i in range(0,x.shape[0],1):
                    for j in range(0,x.shape[1],1):
                        args_expanded += [x[i,j]]
            else:
                raise TypeError('Invalid argument')
        return args_expanded

    # Method: _upgrade_scalars
    def _upgrade_scalars(self,args):
        """
        Upgrades scalars to cvxpy_obj.

        :param args: List of objects.
        """

        args_upgraded = []
        for x in args:
            if np.isscalar(x):
                args_upgraded += [cvxpy_obj(CONSTANT,x,str(x))]
            else:
                args_upgraded += [x]
        return args_upgraded

    # Method: _get_expanded_program
    def _get_expanded_program(self):
        """
        Constructs a new program by applying the
        expansion algorithm to the objective
        and constraints.
        """

        new_obj,obj_constr = expand(self.objective)
        more_constr = expand(self.constraints)
        return cvxpy_program(self.action,new_obj,more_constr+obj_constr,
                             [],self.options,self.name)

    # Method: __getattribute__
    def __getattribute__(self,name):

        # Variables
        if name == 'variables':
            return cvxpy_list(set(self.constraints.variables +
                                  self.objective.variables))

        # Parameters
        elif name == 'parameters':
            return cvxpy_list(set(self.constraints.parameters +
                                  self.objective.parameters))

        # Other
        else:
            return object.__getattribute__(self,name)

    # Method: solve
    def solve(self,quiet=False,return_status=False):
        """
        Solves optimization program using current
        parameter values.
        """

        # Check DCP
        if not self.is_dcp():
            raise ValueError('Program is not DCP')

        # Create param-value map
        replace_map = {}
        for param in self.parameters:
            if np.isnan(param.value):
                raise ValueError('Invalid parameter value: NaN')
            else:
                replace_map[param] = param.value

        # Create new program
        new_p = cvxpy_program(self.action,
                              re_eval(self.objective,replace_map),
                              re_eval(self.constraints,replace_map),
                              [],self.options,'')

        # Solve new program
        obj,valid = solve_prog(new_p,quiet)
        if return_status:
            return obj,valid
        else:
            return obj

    # Method: __str__
    def __str__(self):

        return self.name

    # Method: show
    def show(self):
        """
        Shows description of the optimization program.
        """

        if self.action == MINIMIZE:
            output = '\nminimize '
        else:
            output = '\nmaximize '
        output += str(self.objective)+'\n'
        output += 'subject to\n'
        output += str(self.constraints)
        print(output)

    # Method: __call__
    def __call__(self,*args):
        """
        Invokes program.

        :param args: List of arguments.
        """

        # Get arguments reference
        args_ref = self._get_expanded_objects(self.formals)

        # Internal call
        if len(args) == 1 and type(args[0]) is list:

            # Get list of arguments
            args_list = args[0]

            # Check number of arguments
            if len(args_list) != len(args_ref):
                raise TypeError('Invalid number of arguments')

        # User call
        else:

            # Check number of arguments
            if len(args) != len(self.formals):
                raise TypeError('Invalid number of arguments')

            # Check syntax
            for i in range(0,len(args),1):
                if (np.isscalar(args[i]) or
                    type(args[i]).__name__ in SCALAR_OBJS):
                    if self.formals[i].shape != (1,1):
                        raise ValueError('Invalid argument shape')
                elif (type(args[i]) is cvxpy_matrix or
                      type(args[i]).__name__ in ARRAY_OBJS):
                    if self.formals[i].shape != args[i].shape:
                        raise ValueError('Invalid argument shape')
                else:
                    raise TypeError('Invalid argument')

            # Expand arguments
            args_list = self._get_expanded_objects(args)

        # Verify values
        for i in range(0,len(args_ref)):
            if np.isscalar(args_list[i]) and np.isnan(args_list[i]):
                raise ValueError('Invalid argument value: NaN')

        # Upgrade scalars to objects
        args_upgraded = self._upgrade_scalars(args_list)

        # Verify replacements
        for i in range(0,len(args_ref)):
            if (len(args_upgraded[i].variables) != 0 and
                type(args_ref[i]) is not cvxpy_scalar_var):
                raise TypeError('Invalid replacement')

        # All numeric
        if all(list(map(lambda x: type(x) is cvxpy_obj,args_upgraded))):

            # Create map and new program
            replace_map = {}
            for i in range(0,len(args_ref),1):
                replace_map[args_ref[i]] = args_upgraded[i]
            new_p = cvxpy_program(self.action,
                                  re_eval(self.objective,replace_map),
                                  re_eval(self.constraints,replace_map),
                                  [],self.options,'')

            # Solve
            obj,valid = new_p.solve(quiet=True,return_status=True)
            if not valid:
                raise ValueError('Unable to compute value')
            else:
                return obj

        # Not all numeric
        else:

            # Return expression tree
            return cvxpy_tree(self,args_upgraded)

    # Method: is_dcp
    def is_dcp(self,args=None):
        """
        Determines if program is DCP-compliant.

        :param args: List of scalar arguments.
        """

        # No arguments: Check body of program
        if args == None:
            if (self.action == MINIMIZE and
                not self.objective.is_convex()):
                return False
            elif (self.action == MAXIMIZE and
                  not self.objective.is_concave()):
                return False
            else:
                return self.constraints.is_dcp()

        # Arguments given: Replace arguments and then check
        else:

            # Create reference list of arguments
            args_ref = self._get_expanded_objects(self.formals)

            # Check number of arguments
            if len(args) != len(args_ref):
                raise TypeError('Invalid number of arguments')

            # Create map and new program
            replace_map = {}
            for k in range(0,len(args),1):
                replace_map[args_ref[k]] = args[k]
            new_p = cvxpy_program(self.action,
                                  re_eval(self.objective,replace_map),
                                  re_eval(self.constraints,replace_map),
                                  [],self.options,'')

            # Check dcp on resulting program
            return new_p.is_dcp()

    # Method: _pm_expand
    def _pm_expand(self,constr):
        """
        Returns constraints that embed the program into
        the parent program.

        :param constr: cvxpy_constr of the form
                       self(args), comparison-type, variable.
        """

        # Get arguments
        args = constr.left.children

        # Create argument reference list
        args_ref = self._get_expanded_objects(self.formals)

        # Check number of arguments
        if len(args) != len(args_ref):
            raise TypeError("Invalid number of arguments")

        # Create map and new program
        replace_map = {}
        for k in range(0,len(args),1):
            replace_map[args_ref[k]] = args[k]
        new_p = cvxpy_program(self.action,
                              re_eval(self.objective,replace_map),
                              re_eval(self.constraints,replace_map),
                              [],self.options,'')

        # Embed
        right = constr.right
        new_constr = []
        if self.action == MINIMIZE:
            new_constr += [compare(new_p.objective,LESS_EQUALS,right)]
        else:
            new_constr += [compare(new_p.objective,GREATER_EQUALS,right)]
        new_constr += new_p.constraints

        # Return constraints
        return new_constr



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Entry point for DELSIG-like Delta-Sigma NTF synthesis function
==============================================================
"""

import numpy as np
from warnings import warn
from ..exceptions import PyDsmApproximationWarning
from ._synthesizeNTF0 import synthesizeNTF0
from ._synthesizeNTF1 import synthesizeNTF1
from ..utilities import digested_options

__all__ = ["synthesizeNTF"]


def synthesizeNTF(order=3, osr=64, opt=0, H_inf=1.5, f0=0.0,
                  **options):
    """
    Synthesizes an NTF for a DS modulator by Schreier's approach.

    Parameters
    ----------
    order : int, optional
        the order of the modulator, defaults to 3
    osr : float, optional
        the oversamping ratio (based on the actual signal bandwidth)
    opt : int or list of floats, optional
        flag for optimized zeros, defaults to 0

        * 0 -> not optimized,
        * 1 -> optimized,
        * 2 -> optimized with at least one zero at band-center,
        * 3 -> optimized zeros (with optimizer)
        * 4 -> same as 3, but with at least one zero at band-center
        * [z] -> zero locations in complex form

    H_inf : real, optional
        max allowed peak value of the NTF. Defaults to 1.5
    f0 : real, optional
        center frequency for BP modulators, or 0 for LP modulators.
        Defaults to 0.
        1 corresponds to the sampling frequency, so that 0.5 is the
        maximum value. Value 0 specifies an LP modulator.

    Returns
    -------
    ntf : tuple
        noise transfer function in zpk form.

    Other Parameters
    ----------------
    use_optimizer : bool
        Use: True for for optimizing the zeros with a fast optimization code,
        False otherwise. Defaults can be set by changing the function
        ``default_options`` attribute.

    Raises
    ------
    ValueError
        'Frequency f0 must be less than 0.5' if f0 is out of range

        'Order must be even for a bandpass modulator' if the order is
        incompatible with the modulator type.

        'The opt vector must be of length xxx' if opt is used to explicitly
        pass the NTF zeros and these are in the wrong number.

    RuntimeError
        'Cannot synthesize NTF zeros' if the synthesis fails for some
        unspecified reason.

    Warns
    -----
    PyDsmApproximationWarning
        'Creating a lowpass ntf.' if the center frequency is different
        from zero, but so low that a low pass modulator must be designed.

        'Unable to achieve specified H_inf ...' if the desired H_inf
        cannot be achieved.

        'Iteration limit exceeded' if the routine converges too
        slowly.

    Notes
    -----
    This is actually a wrapper function which calls the appropriate version
    of synthesizeNTF, based on the control flag `optimize_NTF` which
    determines whether to use optimization tools.

    Parameter H_inf is used to enforce the Lee stability criterion.

    If osr or H_inf are low, then the NTF is non optimal. Use
    synthesizeChebyshevNTF instead.
    """
    # Manage options
    opts = digested_options(options, synthesizeNTF.default_options,
                            ['use_optimizer'])
    use_optimizer = opts['use_optimizer']
    # Do the computation
    if f0 > 0.5:
        raise ValueError('Frequency f0 must be less than 0.5')
    if f0 != 0 and f0 < 0.25/osr:
        warn('Creating a lowpass ntf.', PyDsmApproximationWarning)
        f0 = 0
    if f0 != 0 and order % 2 != 0:
        raise ValueError('Order must be even for a bandpass modulator')
    opt = np.asarray(opt)
    if opt.ndim > 1 or (opt.ndim == 1 and opt.size != order):
        raise ValueError('The opt vector must be of length %d' % order)

    if not use_optimizer:
        ntf = synthesizeNTF0(order, osr, opt, H_inf, f0)
    else:
        ntf = synthesizeNTF1(order, osr, opt, H_inf, f0)
    return ntf

synthesizeNTF.default_options = {'use_optimizer': True}



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
Generic utility functions (:mod:`pydsm.utilities`)
==================================================

This module includes some generic utility functions. The stability of these
interfaces may not be guaranteed.

.. currentmodule:: pydsm.utilities


Constants
---------

.. autodata:: EPS
   :annotation:


Functions
---------

.. autosummary::
   :toctree: generated/

   is_negligible     -- Check if a number is close to zero
   chop              -- Chop to zero numbers that are close to zero
   mdot              -- Dot product taking multiple arguments
   digested_options  -- Helper function for the management of default options

Deprecated functions
--------------------

.. autosummary::
   :toctree: generated/

   db        -- Alias for :func:`relab.db`
   cplxpair  -- Alias for :func:`relab.cplxpair`
"""

from __future__ import division, print_function

import numpy as np
from warnings import warn
from .exceptions import PyDsmDeprecationWarning
from . import relab

import sys
if sys.version_info >= (3,):
    from functools import reduce

__all__ = ["is_negligible", "chop", "db", "cplxpair", "mdot", "EPS",
           "digested_options"]


EPS = np.finfo(float).eps
"""The value returned by ``np.finfo(float).eps``, namely the smallest
   representable positive numebr such that ``1.0 + EPS != 1``."""


def is_negligible(x, tol=100*EPS):
    """
    Check if a number is close to zero.

    Parameters
    ----------
    x : float, complex or array_like
        number to be checked
    tol : float, optional
        absolute tolerance. Defaults to 100 times the system epsilon.

    Returns
    -------
    y : bool or array_like of bools
        whether the input number is really close to zero or not.
    """
    return abs(np.asarray(x)) < tol


def chop(x, tol=100*EPS):
    """
    Chop to zero input numbers that are close to zero.

    Parameters
    ----------
    x : float, complex or array_like
        number to process
    tol : float, optional
        absolute tolerance. Defaults to 100 times the system epsilon.

    Returns
    -------
    y : float
        y is zero if x is close to zero according to the tolerance.
        Alternatively, y is x.

    Notes
    -----
    If the input is an array, it is always copied.
    See also `is_negligible`.
    """
    x = np.asarray(x)
    if np.iscomplexobj(x):
        return chop(x.real, tol) + 1j*chop(x.imag, tol)
    y = np.copy(x)
    y[is_negligible(y, tol)] = 0.
    return y


def mdot(*args):
    """
    Dot product taking multiple arguments.

    Parameters
    ----------
    x1, ..., xn : array_like

    Returns
    -------
    y : ndarray
        y=np.dot(np.dot(np.dot(...np.dot(x1,x2),xn_2)xn_1)xn)

    Notes
    -----
    This function is not necessary with recent numpy version where
    one can write things like ``A.dot(B).dot(C)``
    """
    return reduce(np.dot, args)


def digested_options(opts, defaults, keys=[], multikeys=[], emptycheck=True):
    """
    Helper function for the management of default options

    This is used to validate and pre-digest the parameters of functions
    taking a variable number of keyword arguments with the ``**kwargs``
    syntax.

    Parameters
    ----------
    opts : dict
        Dictionary of optional keyword arguments
    defaults : dict
        Dictionary of defaul values
    keys : list
        List of keys to extract whose value can be directly updated from the
        defaults
    multikeys : list
        List of keys to extract whose value must be recursively updated from
        the defaults dictionary (namely, value of a multikey is a dictionary
        itself)
    emptychec : boolean
        Whether to rise an error if items remain in opts after processing.

    Returns
    -------
    eopts : dict
        Dictionary of extracted and validated options. This dictionary
        contains the values from the defaults dictionary, updated with
        the values passed in opts.

    Notes
    -----
    On exit, the input dictionary opt is changed, with all the validated
    options removed from it.
    """
    out = {}
    for key in keys:
        out[key] = opts.pop(key, defaults[key])
    for key in multikeys:
        d = defaults[key].copy()
        d.update(opts.pop(key, []))
        out[key] = d
    if emptycheck and (len(opts) != 0):
        raise TypeError("Unexpected keyword argument(s) '%s'" %
                        list(opts.keys()))
    return out


# Following two functions are deprecated

def db(x, signal_type='voltage', R=1):
    """
    Alias for :func:`pydsm.relab.db`.

     .. deprecated:: 0.11.0
        Function :func:`db` moved to :mod:`pydsm.relab` module
    """
    warn('Function db moved to relab module', PyDsmDeprecationWarning)
    return relab.db(x, signal_type, R)


def cplxpair(x, tol=100*EPS):
    """
    Alias for :func:`pydsm.relab.cplxpair`.

     .. deprecated:: 0.11.0
        Function :func:`cplxpair` moved to :mod:`pydsm.relab` module
    """
    warn('Function cplxpair moved to relab module',
         PyDsmDeprecationWarning)
    return relab.cplxpair(x, tol)



# -*- coding: utf-8 -*-

# Copyright (c) 2013, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
ISO 226 loudness countours (:mod:`pydsm.iso226`)
================================================

Loudness contours from ISO 226.

Contours are returned both as tables of data and as contour functions

.. currentmodule:: pydsm.iso226


Functions returning ISO 226 contours
------------------------------------

.. autosummary::
   :toctree: generated/

   iso226_spl_contour -- Equal loudness contour (tabled)
   iso226_spl_itpl  -- Interpolated equal loudness contour


Functions computing loudness/acoustic pressure
----------------------------------------------

.. autosummary::
   :toctree: generated/

   tabled_L_p  -- Return table of sound pressure levels for given loudness
   tabled_L_N  -- Return table of perceived loudness for given sound pressure


Functions returning data tabled in the standard
-----------------------------------------------

.. autosummary::
   :toctree: generated/

   tabled_f  -- Return table of frequencies in ISO 226
   tabled_alpha_f  -- Return table of exponents for loudness perception
   tabled_L_U  -- Return table of magnitudes of the linear transfer function
   tabled_T_f  -- Return table of thresholds of hearing


Notes
-----
This module uses data from the latest revision of ISO 226 [1]_.
For reference, also consider [2]_.

The ISO standard provides the equal loudness contours as tabled data.
Tables end at 12.5 kHz. Above this frequency equal-loudness-level data are
relatively scarce and tend to be variable [3]_. Yet, it is known
that the human ear has a precipitous decline in sensitivity with increasing
frequency above 15 kHz, to the point that at about 20 kHz the percieved sound
becomes negligible (> 100 dB attenuation) [4]_. For this reason,
this module includes the possibility of delivering some modified ISO contours
where the tabled data is augmented by creating a new data point at 20 kHz
where the behavior found at 20 Hz is replicated.

.. [1] ISO 226:2003 "Acoustics - Normal equal-loudness-level contours"
.. [2] Jeff Tackett, "ISO 226 Equal-Loudness-Level Contour Signal,"
   2005
   (https://www.mathworks.com/matlabcentral/fileexchange/7028)
.. [3] Yoiti Suzuki et al, "Precise and Full-range Determination of
   Two-dimensional Equal Loudness Contours," 2003
   (http://www.mp3-tech.org/programmer/docs/IS-01Y-E.pdf)
.. [4] Robert A. Wannamaker "Psychoacoustically Optimal Noise
   Shaping", J. Audio Eng. Soc., Vol. 40, N. 7/8, 1992 July/August
"""

from __future__ import division, print_function

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline

__all__ = ["tabled_f", "tabled_alpha_f", "tabled_L_U", "tabled_T_f",
           "tabled_L_p", "tabled_L_N",
           "iso226_spl_contour", "iso226_spl_itpl"]

# Tabled ISO 226 parameters
tbl_f = np.asarray(
    [20, 25, 31.5, 40, 50, 63, 80, 100, 125, 160, 200, 250, 315, 400,
     500, 630, 800, 1000, 1250, 1600, 2000, 2500, 3150, 4000, 5000, 6300,
     8000, 10000, 12500])
tbl_alpha_f = np.asarray(
    [0.532, 0.506, 0.480, 0.455, 0.432, 0.409, 0.387, 0.367, 0.349, 0.330,
     0.315, 0.301, 0.288, 0.276, 0.267, 0.259, 0.253, 0.250, 0.246, 0.244,
     0.243, 0.243, 0.243, 0.242, 0.242, 0.245, 0.254, 0.271, 0.301])
tbl_L_U = np.asarray(
    [-31.6, -27.2, -23.0, -19.1, -15.9, -13.0, -10.3, -8.1, -6.2, -4.5,
     -3.1, -2.0, -1.1, -0.4, 0.0, 0.3, 0.5, 0.0, -2.7, -4.1, -1.0, 1.7,
     2.5, 1.2, -2.1, -7.1, -11.2, -10.7, -3.1])
tbl_T_f = np.asarray(
    [78.5, 68.7, 59.5, 51.1, 44.0, 37.5, 31.5, 26.5, 22.1, 17.9, 14.4,
     11.4, 8.6, 6.2, 4.4, 3.0, 2.2, 2.4, 3.5, 1.7, -1.3, -4.2, -6.0, -5.4,
     -1.5, 6.0, 12.6, 13.9, 12.3])


def tabled_f(hfe=False):
    """Table of frequencies in ISO 226.

    Parameters
    ----------
    hfe : bool
        whether the table should be augmented with a data point
        at 20 kHz (High-Frequency-Enhanced table)

    Returns
    -------
    f : array of floats
        the frequency table.
    """
    return np.append(tbl_f, 20E3) if hfe else tbl_f


def tabled_alpha_f(hfe=False):
    """Table of exponents for loudness perception in ISO 226.

    Parameters
    ----------
    hfe : bool
        whether the table should be augmented with a data point
        at 20 kHz (High-Frequency-Enhanced table)

    Returns
    -------
    alpha_f : array of floats
        the exponents table.
    """
    return np.append(tbl_alpha_f, tbl_alpha_f[0]) if hfe else tbl_alpha_f


def tabled_L_U(hfe=False):
    """Table of magnitudes of the linear transfer function in ISO 226.

    Parameters
    ----------
    hfe : bool
        whether the table should be augmented with a data point
        at 20 kHz (High-Frequency-Enhanced table)

    Returns
    -------
    L_U : array of floats
        the magnitudes table.

    Notes
    -----
    The returned values are the magnitude of the linear transfer function
    normalized at 1 kHz.
    """
    return np.append(tbl_L_U, tbl_L_U[0]) if hfe else tbl_L_U


def tabled_T_f(hfe=False):
    """Table of thresholds of hearing in ISO 226.

    Parameters
    ----------
    hfe : bool
        whether the table should be augmented with a data point
        at 20 kHz (High-Frequency-Enhanced table)

    Returns
    -------
    T_f : array of floats
        the thresholds table.
    """
    return np.append(tbl_T_f, tbl_T_f[0]) if hfe else tbl_T_f


# Check that it works fine when L_N is array
def tabled_A_f(L_N, hfe=False):
    """Table of A_f values for given loundess in ISO 226.

    Parameters
    ----------
    L_N : float
        percieved loudness level in phons
    hfe : bool
        whether the table should be augmented with a data point
        at 20 kHz (High-Frequency-Enhanced table)

    Returns
    -------
    A_f : array of floats
        the A_f table.

    Notes
    -----
    1 phon is 1 dB_SPL (sound pressure level) at 1 kHz. Sound pressure levels
    are measured in dBs by referring to a reference pressure level P0 (close
    to the hearing threshold at 1 kHz and set to 20 uPa RMS).
    """
    A_f = (4.47E-3*(10.0**(0.025*L_N)-1.15) +
           (0.4*10.0**((tbl_T_f+tbl_L_U)/10.0-9.0))**tbl_alpha_f)
    return np.append(A_f, A_f[0]) if hfe else A_f


# Check that it works fine when L_N is array
def tabled_L_p(L_N, hfe=False):
    """Table of sound pressure levels for given loudness in ISO 226.

    This function returns a table according to ISO 226 sect 4.1.

    Parameters
    ----------
    L_N : float
        percieved loudness level in phons
    hfe : bool
        whether the table should be augmented with a data point
        at 20 kHz (High-Frequency-Enhanced table)

    Returns
    -------
    L_p : array of floats
        the sound pressure level table. Sound pressure levels are returned
        in DB_SPL

    Notes
    -----
    1 phon is 1 dB_SPL (sound pressure level) at 1 kHz. Sound pressure levels
    are measured in dBs by referring to a reference pressure level P0 (close
    to the hearing threshold at 1 kHz and set to 20 uPa RMS).
    """
    L_p = (10.0/tbl_alpha_f)*np.log10(tabled_A_f(L_N))-tbl_L_U + 94.0
    return np.append(L_p, L_p[0]) if hfe else L_p


# Check that it works fine when L_P is array
def tabled_B_f(L_p, hfe=False):
    """Table of B_f values for given sound pressure in ISO 226.

    Parameters
    ----------
    L_p : float
        sound pressure level in dB_SPL
    hfe : bool
        whether the table should be augmented with a data point
        at 20 kHz (High-Frequency-Enhanced table)

    Returns
    -------
    B_f : array of floats
        the B_f table.

    Notes
    -----
    Sound pressure levels are measured in dBs by referring to a
    reference pressure level P0 (close to the hearing threshold at 1 kHz and
    set to 20 uPa RMS).
    """
    B_f = ((0.4*10**(L_p+tbl_L_U)/10.-9.)**tbl_alpha_f -
           (0.4*10**(tbl_T_f+tbl_L_U)/10.-9.)**tbl_alpha_f +
           0.005135)
    return np.append(B_f, B_f[0]) if hfe else B_f


# Check that it works fine when L_N is array
def tabled_L_N(L_p, hfe=False):
    """Table of perceived loudness levels for given sound pressure in ISO 226.

    This function returns a table according to ISO 226 sect 4.2.

    Parameters
    ----------
    L_p : float
        sound pressure level in dB_SPL
    hfe : bool
        whether the table should be augmented with a data point
        at 20 kHz (High-Frequency-Enhanced table)

    Returns
    -------
    L_N : array of floats
        the perceived loudness level table. Loudness levels are returned
        in phons

    Notes
    -----
    1 phon is 1 dB_SPL (sound pressure level) at 1 kHz. Sound pressure levels
    are measured in dBs by referring to a reference pressure level P0 (close
    to the hearing threshold at 1 kHz and set to 20 uPa RMS).
    """
    L_N = 40*np.log10(tabled_B_f(L_p))+94.0
    return np.append(L_N, L_N[0]) if hfe else L_N


def iso226_spl_contour(L_N=40, hfe=False):
    """Generates an equal loudness contour as described in ISO 226.

    This function returns the control points describing the equal loudness
    contour for the input phon level, according to ISO 226 sect 4.1.

    Parameters
    ----------
    L_N : float, optional
        perceived loudness level in phons.
    hfe : bool
        whether the table should be augmented with a data point
        at 20 kHz (High-Frequency-Enhanced table)

    Returns
    -------
    f : ndarray
        frequencies where the loudness is evaluated.
    spl : ndarray
        equivalent sound pressure level at the frequencies f.

    Notes
    -----
    1 phon is 1 dB_SPL (sound pressure level) at 1 kHz. Sound pressure levels
    are measured in dBs by referring to a reference pressure level P0 (close
    to the hearing threshold at 1 kHz and set to 20 uPa RMS).

    The valid input phon range is 0-90 dB_SPL. Above 80 dB, only the
    frequency range 20-4000 Hz is significant.
    """
    # Check for valid input range
    if L_N < 0 or L_N > 90:
        raise ValueError('Parameter L_N out of bounds [0-90].')
    # Derive sound pressure level from loudness level (ISO 226 sect 4.1)
    return tabled_f(hfe), tabled_L_p(L_N, hfe)


def iso226_spl_itpl(L_N=40, hfe=False, k=3):
    """Generates an interpolation of an equal loudness contour.

    This function returns an interpolation object describing the equal
    loudness contour for the input phon level, according to ISO 226 sect 4.1.

    Parameters
    ----------
    L_N : float, optional
        perceived loudness level in phons.
    hfe : bool
        whether the table should be augmented with a data point
        at 20 kHz (High-Frequency-Enhanced table)
    k : int
        interpolation order

    Returns
    -------
    itpl : univariate interpolation object
        function-like object that takes a frequency f as its input and returns
        the equivalent sound pressure level at f

    Notes
    -----
    1 phon is 1 dB_SPL (sound pressure level) at 1 kHz. Sound pressure levels
    are measured in dBs by referring to a reference pressure level P0 (close
    to the hearing threshold at 1 kHz and set to 20 uPa RMS).

    The valid input phon range is 0-90 dB_SPL. Above 80 dB, only the
    frequency range 20-4000 Hz is significant.
    """
    ff, yy = iso226_spl_contour(L_N, hfe)
    return InterpolatedUnivariateSpline(ff, yy, k=k)



# Copyright © 2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

def pytest_configure(config):
    config.addinivalue_line("python_files", "benchmark_*.py")
    config.addinivalue_line("python_functions", "benchmark_*")
    config.addinivalue_line("python_classes", "Benchmark_*")
    config.addinivalue_line("markers",
        "slow: Tests that are very slow.")



# -*- coding: utf-8 -*-

# Copyright (c) 2013–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import picos
import cvxopt


def ntf_fir_from_digested(Qs, A, C, H_inf, **opts):
    """
    Synthesize FIR NTF from predigested specification

    Version for the picos modeler.
    """
    verbose = 1 if opts.get('show_progress', True) else 0
    if 'maxiters' in opts['picos_opts']:
        opts['picos_opts']['max_iterations'] = opts['picos_opts']['maxiters']
        del opts['picos_opts']['maxiters']
    opts['picos_opts']['abs_ipm_opt_tol']=opts['picos_opts']['abstol']
    del opts['picos_opts']['abstol']
    opts['picos_opts']['rel_ipm_opt_tol']=opts['picos_opts']['reltol']
    del opts['picos_opts']['reltol']
    del opts['picos_opts']['feastol']
    # Do the computation
    order = int(np.size(Qs, 0)-1)
    Qs = cvxopt.matrix(Qs)
    p = picos.Problem(solver='cvxopt')
    power = picos.RealVariable('pow')
    br = picos.RealVariable('br', (order, 1))
    b = 1 // br
    X = picos.SymmetricVariable('X', order)
    A = cvxopt.matrix(A)
    B = cvxopt.matrix(np.vstack((np.zeros((order-1, 1)), 1.)))
    C = cvxopt.matrix(C)+br[::-1, :].T
    D = cvxopt.matrix(1.)
    M1 = A.T*X
    M2 = M1*B
    M = (((M1*A-X) & M2 & C.T) //
         (M2.T & (B.T*X*B-H_inf**2) & D) //
         (C & D & -1.))
    constraint1 = (M << 0)
    constraint2 = (X >> 0)
    p.set_objective('min', power)
    p.add_constraint(abs(Qs*b) <= power)
    p.add_constraint(constraint1)
    p.add_constraint(constraint2)
    p.solve(verbosity=verbose, **opts['picos_opts'])
    return np.hstack((1, np.asarray(br.value.T)[0]))



# -*- coding: utf-8 -*-

# Copyright (c) 2015–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes portions of code ported from the "Frequency-Domain
# Min-Max Optimization for Delta-Sigma Modulators" Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/36187)
# covered by the following copyright and permission notice
#
# Copyright (c) 2012 Masaaki Nagahara
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from __future__ import division, print_function

import numpy as np
import picos
import cvxopt


def ntf_fir_from_digested(order, osrs, H_inf, f0s, zf, **opts):
    """
    Synthesize FIR NTF with minmax approach from predigested specification

    Version for the PICOS modeler.
    """
    verbose = 1 if opts.get('show_progress', True) else 0
    if 'maxiters' in opts['picos_opts']:
        opts['picos_opts']['max_iterations'] = opts['picos_opts']['maxiters']
        del opts['picos_opts']['maxiters']
    opts['picos_opts']['abs_ipm_opt_tol']=opts['picos_opts']['abstol']
    del opts['picos_opts']['abstol']
    opts['picos_opts']['rel_ipm_opt_tol']=opts['picos_opts']['reltol']
    del opts['picos_opts']['reltol']
    del opts['picos_opts']['feastol']

    # State space representation of NTF
    A = cvxopt.matrix(np.eye(order, order, 1))
    B = cvxopt.matrix(np.vstack((np.zeros((order-1, 1)), 1.)))
    # C contains the NTF coefficients
    D = cvxopt.matrix(1.)

    p = picos.Problem()
    # Set up the problem
    bands = len(f0s)
    c = picos.RealVariable('c', (1, order))
    gg = picos.RealVariable('gg', bands)
    #gg = p.add_variable('gg', bands)

    for idx in range(bands):
        f0 = f0s[idx]
        osr = osrs[idx]
        omega0 = 2*f0*np.pi
        Omega = 1./osr*np.pi
        P = picos.SymmetricVariable(f'P[{idx}]', order)
        Q = picos.SymmetricVariable(f'Q[{idx}]', order)
        if f0 == 0:
            # Lowpass modulator
            M1 = A.T*P*A+Q*A+A.T*Q-P-2*Q*np.cos(Omega)
            M2 = A.T*P*B + Q*B
            M3 = B.T*P*B - gg[idx, 0]
            M = ((M1 & M2 & c.T) //
                 (M2.T & M3 & D) //
                 (c & D & -1.0))
            p.add_constraint(Q >> 0)
            p.add_constraint(M << 0)
            if zf:
                # Force a zero at DC
                p.add_constraint(picos.sum(c[i] for i in range(order)) == -1)
        else:
            # Bandpass modulator
            M1r = (A.T*P*A + Q*A*np.cos(omega0) + A.T*Q*np.cos(omega0) -
                   P - 2*Q*np.cos(Omega))
            M2r = A.T*P*B + Q*B*np.cos(omega0)
            M3r = B.T*P*B - gg[idx]
            M1i = A.T*Q*np.sin(omega0) - Q*A*np.sin(omega0)
            M21i = -Q*B*np.sin(omega0)
            M22i = B.T*Q*np.sin(omega0)
            Mr = ((M1r & M2r & c.T) //
                  (M2r.T & M3r & D) //
                  (c & D & -1.0))
            Mi = ((M1i & M21i & cvxopt.matrix(np.zeros((order, 1)))) //
                  (M22i & 0.0 & 0.0) //
                  (cvxopt.matrix(np.zeros((1, order+2)))))
            M = ((Mr & Mi) //
                 (-Mi & Mr))
            p.add_constraint(Q >> 0.0)
            p.add_constraint(M << 0.0)
            if zf:
                # Force a zero at z=np.exp(1j*omega0)
                nn = np.arange(order).reshape((order, 1))
                vr = cvxopt.matrix(np.cos(omega0*nn))
                vi = cvxopt.matrix(np.sin(omega0*nn))
                vn = cvxopt.matrix(
                    [-np.cos(omega0*order), -np.sin(omega0*order)])
                p.add_constraint(c*(vr & vi) == vn)
    if H_inf < np.inf:
        # Enforce the Lee constraint
        R = picos.SymmetricVariable('R', order)
        p.add_constraint(R >> 0)
        MM = ((A.T*R*A-R & A.T*R*B & c.T) //
              (B.T*R*A & -H_inf**2+B.T*R*B & D) //
              (c & D & -1))
        p.add_constraint(MM << 0)
    mx = picos.RealVariable('mx')
    p.add_constraint(picos.sum_k_largest(gg, 1) <= mx)
    p.set_objective('min', mx)
    p.solve(verbosity=verbose, **opts['picos_opts'])
    return np.hstack((1, np.asarray(c.value)[0, ::-1]))



# -*- coding: utf-8 -*-

# Copyright (c) 2013–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import cvxpy


def ntf_fir_from_digested(Qs, A, C, H_inf, **opts):
    """
    Synthesize FIR NTF from predigested specification

    Version for the cvxpy modeler.
    """
    verbose = opts['show_progress']
    if opts['cvxpy_opts']['solver'] == 'cvxopt':
        opts['cvxpy_opts']['solver'] = cvxpy.CVXOPT
    elif opts['cvxpy_opts']['solver'] == 'scs':
        opts['cvxpy_opts']['solver'] = cvxpy.SCS
    order = int(np.size(Qs, 0)-1)
    br = cvxpy.Variable((order, 1), name='br')
    b = cvxpy.vstack([np.array([[1]]), br])
    X = cvxpy.Variable((order, order), symmetric=True, name='X')
    target = cvxpy.Minimize(cvxpy.norm2(Qs @ b))
    B = np.vstack((np.zeros((order-1, 1)), 1.))
    C = C+br[::-1].T
    D = np.array([[1.]])
    M1 = A.T @ X
    M2 = M1 @ B
    M = cvxpy.bmat([[M1 @ A-X, M2, C.T],
                    [M2.T, B.T @ X @ B-H_inf**2, D],
                    [C, D, np.array([[-1.]])]])
    constraints = [M << 0, X >> 0]
    p = cvxpy.Problem(target, constraints)
    p.solve(verbose=verbose, **opts['cvxpy_opts'])
    return np.hstack((1, np.asarray(br.value.T)[0]))



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..scalars import *
from ..constraints import *
from ..interface import equals
from ..interface import less_equals
from ..interface import greater_equals
from ..interface import variable
from ..arrays import cvxpy_var
from ..arrays import cvxpy_array

# Function: expand
def expand(arg):

    # Constant
    if type(arg) is cvxpy_obj:
        return arg,[]

    # Scalar variable
    elif type(arg) is cvxpy_scalar_var:
        return arg,[]

    # Summation
    elif (type(arg) is cvxpy_tree and
          arg.item.type == OPERATOR and
          arg.item.name == SUMMATION):

        # Get item and children
        item = arg.item
        children = arg.children

        # New variale
        v = variable()

        # Expand children
        new_children = []
        new_constr = []
        for child in children:

            # Multiplication
            if (child.type == TREE and
                child.item.type == OPERATOR and
                child.item.name == MULTIPLICATION):
                child_var,child_constr = expand(child.children[1])
                new_children += [child.children[0].value*child_var]
                new_constr += child_constr

            # Else
            else:
                child_var,child_constr = expand(child)
                new_children += [child_var]
                new_constr += child_constr

        # Return (Always right side is the new variable)
        new_tree = cvxpy_tree(item,new_children)
        return v,[equals(new_tree,v)]+new_constr

    # Multiplication
    elif (type(arg) is cvxpy_tree and
          arg.item.type == OPERATOR and
          arg.item.name == MULTIPLICATION):

        # Get item and children
        item = arg.item
        children = arg.children

        # New variable
        v = variable()

        # Apply expand to second operand (first is a constant)
        child_var,child_constr = expand(children[1])

        # Return result (Always right side is the new variable)
        new_tree = cvxpy_tree(item,[children[0],child_var])
        new_eq = [equals(new_tree,v)]
        new_eq += child_constr
        return v,new_eq

    # Function
    elif (type(arg) is cvxpy_tree and
          arg.item.type == FUNCTION):

        # Get item and children
        item = arg.item
        children = arg.children

        # New varibale
        v = variable()

        # Analyze children
        new_children = []
        new_constr = []
        for child in children:
            child_var,child_constr = expand(child)
            new_children += [child_var]
            new_constr += child_constr

        # Return (Always right side is the new variable)
        new_tree = cvxpy_tree(item,new_children)
        if arg.is_convex():
            new_constr += [less_equals(new_tree,v)]
        else:
            new_constr += [greater_equals(new_tree,v)]
        return v,new_constr

    # Constraint
    elif type(arg) is cvxpy_constr:

        # Not set membership
        if arg.type != BELONGS:

            # Apply expand to left and right side
            obj1,constr_list1 = expand(arg.left)
            obj2,constr_list2 = expand(arg.right)

            # Return new constraints
            new_constr = cvxpy_constr(obj1,arg.type,obj2)
            new_list = [new_constr]
            new_list += constr_list1
            new_list += constr_list2
            return new_list

        # Set membership
        else:
            obj,constr_list = expand(arg.left)
            new_constr = cvxpy_constr(obj,arg.type,arg.right)
            return [new_constr]+constr_list

    # Array
    elif (type(arg) is cvxpy_array or
          type(arg) is cvxpy_var):
        (m,n) = arg.shape
        new_list = []
        new_ar = cvxpy_array(m,n)
        for i in range(0,m,1):
            for j in range(0,n,1):

                # Number: Upgrade
                if np.isscalar(arg[i,j]):
                    new_ar[i,j] = cvxpy_obj(CONSTANT,arg[i,j],str(arg[i,j]))

                # Not a number
                else:
                    obj,constr_list = expand(arg[i,j])
                    new_ar[i,j] = obj
                    new_list += constr_list
        return new_ar,new_list

    # List of constraints
    elif (type(arg) is list or
          type(arg) is cvxpy_list):
        return reduce(lambda x,y:x+y,list(map(expand,arg)),[])

    # Invalid
    else:
        raise TypeError('Invalid argument')



# -*- coding: utf-8 -*-

# Copyright (c) 2013–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
Design optimal NTF with NTF zeros as parameters

Routines to design an optimal NTF from a weighting function using
the NTF zeros as free parameters.

The optimization is practiced in two steps:

* First a matrix (symmetric, Toeplitz, Positive semidefinite) is
  extracted from the weighting function to define a quadratic cost
  function for the optimization problem. The Toeplitz-symmetric property of
  the matrix let it be fully defined from its first row only.
* Then the NTF zeros are optimized based on this cost function.
"""

from __future__ import division, print_function

import numpy as np
from ...ft import idtft_hermitian
from ...delsig import evalTF, padr
from warnings import warn
from ...exceptions import PyDsmDeprecationWarning
from ...utilities import digested_options
import scipy.linalg as la

__all__ = ["q0_from_noise_weighting", "q0_weighting",
           "ntf_fir_from_q0", "synthesize_ntf_from_q0",
           "ntf_fir_weighting", "synthesize_ntf_from_noise_weighting",
           "mult_weightings", "ntf_hybrid_weighting"]


def mult_weightings(*ww):
    """
    Product of weighting functions.

    Returns a function that is the product of many weighting functions.

    Parameters
    ----------
    w1, w2, ... : functions or tuples
        Weighting functions.
        If an entry is a function, it is used as is.
        If it is a tuple, it interpreted as filters in ba or zpk form from
        which a weighting function is implicitly obtained.

    Returns
    -------
    w : function
        Overall weighting function
    """
    wn = [0] * len(ww)
    for i, wi in enumerate(ww):
        if type(wi) is tuple and 2 <= len(wi) <= 3:
            wn[i] = (lambda wi:
                     lambda f: np.abs(evalTF(wi, np.exp(2j*np.pi*f)))**2)(wi)
        else:
            wn[i] = wi
    return lambda f: np.prod([w(f) for w in wn], axis=0)


def q0_weighting(P, w, **options):
    """Compute Q matrix from a noise weighting function or a filter

    Parameters
    ----------
    P : int
        order of the FIR to be eventually synthesized
    w : callable with argument f in [0,1/2] or tuple
            * if function: noise weighting function
            * if filter definition as zpk or ba tuple: weighting is implicitly
              provided by the filter

    Returns
    -------
    q0 : ndarray
        the first row of the matrix Q used in the NTF optimization

    Other parameters
    ----------------
    quad_opts : dictionary, optional
        Parameters to be passed to the ``quad`` function used internally as
        an integrator. Allowed options are ``epsabs``, ``epsrel``, ``limit``,
        ``points``. Do not use other options since they could break the
        integrator in unexpected ways.

    Notes
    -----
    The Q matrix being synthesized has (P+1) times (P+1) entries.

    Default values for the options not directly documented in the function
    call signature can be checked and updated by changing the function
    ``default_options`` attribute.

    See Also
    --------
    scipy.integrate.quad : For the meaning of the integrator parameters.
    """
    # Manage parameters
    if type(w) is tuple and 2 <= len(w) <= 3:
        h = w
        w = lambda f: np.abs(evalTF(h, np.exp(2j*np.pi*f)))**2
    # Manage optional parameters
    opts = digested_options(options, q0_weighting.default_options,
                            [], ['quad_opts'])
    # Do the computation
    return idtft_hermitian(w, np.arange(P+1), **opts)

q0_weighting.default_options = {"quad_opts": {"epsabs": 1E-14,
                                              "epsrel": 1E-9,
                                              "limit": 100,
                                              "points": None}}


def ntf_fir_from_q0(q0, H_inf=1.5, normalize="auto", **options):
    """Synthesize FIR NTF from quadratic form expressing noise weighting.

    Parameters
    ----------
    q0 : ndarray
        first row of the Toeplitz symmetric matrix defining the quadratic form
    H_inf : real, optional
        Max peak NTF gain, defaults to 1.5, used to enforce the Lee criterion
    normalize : string or real, optional
        Normalization to apply to the quadratic form used in the NTF
        selection. Defaults to 'auto' which means setting the top left entry
        in the matrix Q defining the quadratic form to 1.

    Returns
    -------
    ntf : ndarray
        FIR NTF in zpk form

    Other parameters
    ----------------
    show_progress : bool, optional
        provide extended output.
    fix_pos : bool, optional
        fix quadratic form for positive definiteness. Numerical noise
        may make it not positive definite leading to errors.
    modeler : string, optional
        modeling backend for the optimization problem. Currently, the
        ``cvxpy_old``, ``cvxpy`` and ``picos`` backends are supported.
        Default is ``cvxpy_old``.
    cvxpy_opts : dictionary, optional
       A dictionary of options to use with the ``cvxpy`` modeling library.
       Allowed options include:

       ``override_kktsolver`` (bool)
           Whether to override the default ``cvxopt`` kkt solver using the
           ``chol`` kkt solver.
           Leave this at the default True setting, to avoid paying a
           performance price.
       ``solver`` (string)
           The solver backend to use. Either `cvxopt` or `scs`

    cvxopt_opts : dict, optional
        A dictionary of options for the ``cvxopt`` optimizer.
        Allowed options include:

        ``maxiters`` (int)
            Maximum number of iterations
        ``abstol`` (real)
            Absolute accuracy
        ``reltol`` (real)
            Relative accuracy
        ``feastol`` (real)
            Tolerance for feasibility conditions (defaults to 1e-6)

        Do not use other options since they could break ``CVXOPT`` in
        unexpected ways. These options can be passed when using the
        ``cvxpy_old`` modeler, the ``picos`` modeler or the ``cvxpy`` modeler
        with the ``cvxopt`` backend.

    scs_opts : dict, optional
        A dictionary of options for the ``scs`` optimizer.  Allowed options
        include:

        ``max_iters`` (int)
            Maximum number of iterations
        ``eps`` (real)
            Convergence tolerance
        ``alpha`` (real)
            Relaxation parameter
        ``normalize`` (bool)
            Whether to precondition data matrices
        ``use_indirect`` (bool)
            Whether to use indirect solver for KKT sytem (instead of direct)

       Do not use other options since they could break ``scs`` in
       unexpected ways. These options can be passed when using the
       ``cvxpy`` modeler with the ``scs`` backend.

    Notes
    -----
    Default values for the options not directly documented in the
    function call signature can be checked and updated by changing the
    function ``default_options`` attribute. For more information on
    the specific options taken by the ``CVXOPT`` and ``SCS``
    optimizer, look at the corresponding documentation. Similarly, for
    more information about the ``cvxpy`` modeler and optimization
    frontend, look at the ``cvxpy`` documentation.
    """
    # Manage optional parameters
    opts = digested_options(
        options, ntf_fir_from_q0.default_options,
        ['show_progress', 'fix_pos', 'modeler'], [], False)
    dig_opts = {'show_progress': opts['show_progress'],
                'cvxpy_opts': {},
                'cvxpy_tdr_opts': {},
                'picos_opts': {}}
    if opts['modeler'] == 'cvxpy':
        opts.update(digested_options(
            options, ntf_fir_from_q0.default_options,
            [], ['cvxpy_opts'], False))
        if opts['cvxpy_opts']['solver'] == 'cvxopt':
            dig_opts['cvxpy_opts'].update(digested_options(
                options, ntf_fir_from_q0.default_options,
                [], ['cvxopt_opts'], False)['cvxopt_opts'])
            if opts['cvxpy_opts'].get('override_kktsolver', True):
                dig_opts['cvxpy_opts']['kktsolver'] = 'chol'
        elif opts['cvxpy_opts']['solver'] == 'scs':
            dig_opts['cvxpy_opts'].update(digested_options(
                options, ntf_fir_from_q0.default_options,
                [], ['scs_opts'], False)['scs_opts'])
        opts['cvxpy_opts'].pop('override_kktsolver')
        dig_opts['cvxpy_opts'].update(opts['cvxpy_opts'])
        from ._fir_weighting_cvxpy import (
            ntf_fir_from_digested as _ntf_fir_from_digested)
    elif opts['modeler'] == 'cvxpy_old':
        dig_opts['cvxpy_tdr_opts'].update(digested_options(
            options, ntf_fir_from_q0.default_options,
            [], ['cvxopt_opts'], False)['cvxopt_opts'])
        from ._fir_weighting_cvxpy_tdr import (
            ntf_fir_from_digested as _ntf_fir_from_digested)
    elif opts['modeler'] == 'picos':
        dig_opts['picos_opts'].update(digested_options(
            options, ntf_fir_from_q0.default_options,
            [], ['cvxopt_opts'], False)['cvxopt_opts'])
        from ._fir_weighting_picos import (
            ntf_fir_from_digested as _ntf_fir_from_digested)
    else:
        raise ValueError('Unsupported modeling backend {}'.format(
            opts['modeler']))
    digested_options(options, {})
    # Do the computation
    if normalize == 'auto':
        q0 = q0/q0[0]
    elif normalize is not None:
        q0 = q0*normalize
    order = q0.shape[0]-1
    Q = la.toeplitz(q0)
    d, v = np.linalg.eigh(Q)
    if opts['fix_pos']:
        d = d/np.max(d)
        d[d < 0] = 0.
    Qs = v.dot(np.diag(np.sqrt(d))).dot(v.T)
    A = np.eye(order, order, 1)
    C = np.zeros((1, order))
    ntf_ir = _ntf_fir_from_digested(Qs, A, C, H_inf, **dig_opts)
    return (np.roots(ntf_ir), np.zeros(order), 1.)


ntf_fir_from_q0.default_options = {"modeler": "cvxpy_old",
                                   "cvxpy_opts": {'override_kktsolver': True,
                                                  'solver': 'cvxopt'},
                                   "cvxopt_opts": {'maxiters': 100,
                                                   'abstol': 1e-7,
                                                   'reltol': 1e-6,
                                                   'feastol': 1e-6},
                                   'scs_opts': {'max_iters': 250000,
                                                'eps': 1e-10,
                                                'alpha': 1.8,
                                                'normalize': True,
                                                'use_indirect': False,
                                                'acceleration_lookback': 0},
                                   'show_progress': True,
                                   'fix_pos': True}


def ntf_fir_weighting(order, w, H_inf=1.5,
                      normalize="auto", **options):
    u"""Synthesize FIR NTF based on a noise weighting function or a filter.

    The ΔΣ modulator NTF is designed after a noise weigthing function stating
    how expensive noise is at the various frequencies.

    Parameters
    ----------
    order : int
        Delta sigma modulator order
    w : callable with argument f in [0,1/2] or tuple
            * if function: noise weighting function
            * if filter definition as zpk or ba tuple: weighting is implicitly
              provided by the filter
    H_inf : real, optional
        Max peak NTF gain, defaults to 1.5, used to enforce the Lee criterion
    normalize : string or real, optional
        Normalization to apply to the quadratic form used in the NTF
        selection. Defaults to 'auto' which means setting the top left entry
        in the matrix Q defining the quadratic form to 1.

    Returns
    -------
    ntf : ndarray
        FIR NTF in zpk form

    Other parameters
    ----------------
    show_progress : bool, optional
        provide extended output.
    fix_pos : bool, optional
        fix quadratic form for positive definiteness. Numerical noise
        may make it not positive definite leading to errors.
    modeler : string, optional
        modeling backend for the optimization problem. Currently, the
        ``cvxpy_old``, ``cvxpy`` and ``picos`` backends are supported.
        Default is ``cvxpy_old``.
    cvxpy_opts : dictionary, optional
       A dictionary of options to use with the ``cvxpy`` modeling library.
       Allowed options include:

       ``override_kktsolver`` (bool)
           Whether to override the default ``cvxopt`` kkt solver using the
           ``chol`` kkt solver.
           Leave this at the default True setting, to avoid paying a
           performance price.
       ``solver`` (string)
           The solver backend to use. Either `cvxopt` or `scs`

    cvxopt_opts : dict, optional
        A dictionary of options for the ``cvxopt`` optimizer.
        Allowed options include:

        ``maxiters`` (int)
            Maximum number of iterations
        ``abstol`` (real)
            Absolute accuracy
        ``reltol`` (real)
            Relative accuracy
        ``feastol`` (real)
            Tolerance for feasibility conditions (defaults to 1e-6)

        Do not use other options since they could break ``cvxopt`` in
        unexpected ways. These options can be passed when using the
        ``cvxpy_old`` modeler, the ``picos`` modeler or the ``cvxpy`` modeler
        with the ``cvxopt`` backend.
    scs_opts : dict, optional
        A dictionary of options for the ``scs`` optimizer.  Allowed options
        include:

        ``max_iters`` (int)
            Maximum number of iterations
        ``eps`` (real)
            Convergence tolerance
        ``alpha`` (real)
            Relaxation parameter
        ``normalize`` (bool)
            Whether to precondition data matrices
        ``use_indirect`` (bool)
            Whether to use indirect solver for KKT sytem (instead of direct)

       Do not use other options since they could break ``scs`` in
       unexpected ways. These options can be passed when using the
       ``cvxpy`` modeler with the ``scs`` backend.
    quad_opts : dictionary, optional
        Parameters to be passed to the ``quad`` function used internally as
        an integrator. Allowed options are ``epsabs``, ``epsrel``, ``limit``,
        ``points``. Do not use other options since they could break the
        integrator in unexpected ways.

    Notes
    -----
    Default values for the options not directly documented in the function
    call signature can be checked and updated by changing the function
    ``default_options`` attribute.

    The internal operation of this function is described in [1]_.

    .. [1] Sergio Callegari, Federico Bizzarri *“Noise Weighting in the
       Design of ΔΣ Modulators (with a Psychoacoustic Coder as an
       Example),”* IEEE Transactions on Circuits and Systems - Part II:
       Express Briefs, Vol. 60, N. 11, pp. 756-760. Nov. 2013. DOI:
       `10.1109/TCSII.2013.2281892
       <https://dx.doi.org/10.1109/TCSII.2013.2281892>`__. Pre-print
       available on `arXiv <https://arxiv.org/abs/1309.6151>`__.

    For the specific options taken by the modeler and optimization
    frontend ``cvxpy`` as well as the optimizers ``CVXOPT`` and
    ``scs``, look at the documentation of the individual packages.

    See Also
    --------
    scipy.integrate.quad : for the meaning of the integrator parameters
    """
    # Manage optional parameters
    opts1 = digested_options(options, ntf_fir_weighting.default_options,
                             [], ['quad_opts'], False)
    opts2 = digested_options(
        options, ntf_fir_weighting.default_options,
        ['show_progress', 'fix_pos', 'modeler'], [], False)
    if opts2['modeler'] == 'cvxpy':
        opts2.update(digested_options(
            options, ntf_fir_weighting.default_options,
            [], ['cvxpy_opts'], False))
        if opts2['cvxpy_opts']['solver'] == 'cvxopt':
            opts2.update(digested_options(
                options, ntf_fir_weighting.default_options,
                [], ['cvxopt_opts'], False))
        elif opts2['cvxpy_opts']['solver'] == 'scs':
            opts2.update(digested_options(
                options, ntf_fir_weighting.default_options,
                [], ['scs_opts'], False))
    elif opts2['modeler'] == 'cvxpy_old' or opts2['modeler'] == 'picos':
            opts2.update(digested_options(
                options, ntf_fir_weighting.default_options,
                [], ['cvxopt_opts'], False))
    else:
        raise ValueError('Unsupported modeling backend {}'.format(
            opts2['modeler']))
    digested_options(options, {})
    # Do the computation
    q0 = q0_weighting(order, w, **opts1)
    return ntf_fir_from_q0(q0, H_inf, normalize, **opts2)

ntf_fir_weighting.default_options = q0_weighting.default_options.copy()
ntf_fir_weighting.default_options.update(ntf_fir_from_q0.default_options)


def ntf_hybrid_weighting(order, w, H_inf=1.5, poles=[],
                         normalize="auto", **options):
    u"""
    Synthesize NTF based on noise weighting function or filter plus poles.

    The ΔΣ modulator NTF is designed after a noise weigthing function stating
    how expensive noise is at the various frequencies.

    Parameters
    ----------
    order : int
        Delta sigma modulator order
    w : callable with argument f in [0,1/2] or tuple
            * if function: noise weighting function
            * if filter definition as zpk or ba tuple: weighting is implicitly
              provided by the filter
    H_inf : real, optional
        Max peak NTF gain, defaults to 1.5, used to enforce the Lee criterion
    poles : array_like
        List of pre-assigned NTF poles. Must be no longer than ``order``.
    normalize : string or real, optional
        Normalization to apply to the quadratic form used in the NTF
        selection. Defaults to 'auto' which means setting the top left entry
        in the matrix Q defining the quadratic form to 1.

    Returns
    -------
    ntf : ndarray
        FIR NTF in zpk form

    Other parameters
    ----------------
    show_progress : bool, optional
        provide extended output.
    fix_pos : bool, optional
        fix quadratic form for positive definiteness. Numerical noise
        may make it not positive definite leading to errors.
    modeler : string, optional
        modeling backend for the optimization problem. Currently, the
        ``cvxpy_old``, ``cvxpy`` and ``picos`` backends are supported.
        Default is ``cvxpy_old``.
    cvxpy_opts : dictionary, optional
       A dictionary of options to use with the ``cvxpy`` modeling library.
       Allowed options include:

       ``override_kktsolver`` (bool)
           Whether to override the default ``cvxopt`` kkt solver using the
           ``chol`` kkt solver.
           Leave this at the default True setting, to avoid paying a
           performance price.
       ``solver`` (string)
           The solver backend to use. Either `cvxopt` or `scs`

    cvxopt_opts : dict, optional
        A dictionary of options for the ``cvxopt`` optimizer.
        Allowed options include:

        ``maxiters`` (int)
            Maximum number of iterations
        ``abstol`` (real)
            Absolute accuracy
        ``reltol`` (real)
            Relative accuracy
        ``feastol`` (real)
            Tolerance for feasibility conditions (defaults to 1e-6)

        Do not use other options since they could break ``cvxopt`` in
        unexpected ways. These options can be passed when using the
        ``cvxpy_old`` modeler, the ``picos`` modeler or the ``cvxpy`` modeler
        with the ``cvxopt`` backend.
    scs_opts : dict, optional
        A dictionary of options for the ``scs`` optimizer.  Allowed options
        include:

        ``max_iters`` (int)
            Maximum number of iterations
        ``eps`` (real)
            Convergence tolerance
        ``alpha`` (real)
            Relaxation parameter
        ``normalize`` (bool)
            Whether to precondition data matrices
        ``use_indirect`` (bool)
            Whether to use indirect solver for KKT sytem (instead of direct)

       Do not use other options since they could break ``scs`` in
       unexpected ways. These options can be passed when using the
       ``cvxpy`` modeler with the ``scs`` backend.
    quad_opts : dictionary, optional
        Parameters to be passed to the ``quad`` function used internally as
        an integrator. Allowed options are ``epsabs``, ``epsrel``, ``limit``,
        ``points``. Do not use other options since they could break the
        integrator in unexpected ways.

    Notes
    -----
    Default values for the options not directly documented in the function
    call signature can be checked and updated by changing the function
    ``default_options`` attribute.

    The internal operation of this function is described in [1]_.

    .. [1] Sergio Callegari, Federico Bizzarri "Optimal Design of the Noise
       Transfer Function of ΔΣ Modulators: IIR strategies, FIR strategies,
       FIR strategies with Preassigned Poles," Signal Processing, Elsevier,
       2015. DOI:  10.1016/j.sigpro.2015.02.001.

    For the specific options taken by the modeler and optimization
    frontend ``cvxpy`` as well as the optimizers ``CVXOPT`` and
    ``scs``, look at the documentation of the individual packages.

    See Also
    --------
    scipy.integrate.quad : for the meaning of the integrator parameters
    """
    # Manage optional parameters
    opts1 = digested_options(options, ntf_hybrid_weighting.default_options,
                             [], ['quad_opts'], False)
    opts2 = digested_options(
        options, ntf_hybrid_weighting.default_options,
        ['show_progress', 'fix_pos', 'modeler'], [], False)
    dig_opts = {'show_progress': opts2['show_progress'],
                'cvxpy_opts': {},
                'cvxpy_tdr_opts': {},
                'picos_opts': {}}
    if opts2['modeler'] == 'cvxpy':
        opts2.update(digested_options(
            options, ntf_fir_from_q0.default_options,
            [], ['cvxpy_opts'], False))
        if opts2['cvxpy_opts']['solver'] == 'cvxopt':
            dig_opts['cvxpy_opts'].update(digested_options(
                options, ntf_fir_from_q0.default_options,
                [], ['cvxopt_opts'], False)['cvxopt_opts'])
            if opts2['cvxpy_opts'].get('override_kktsolver', True):
                dig_opts['cvxpy_opts']['kktsolver'] = 'chol'
        elif opts2['cvxpy_opts']['solver'] == 'scs':
            dig_opts['cvxpy_opts'].update(digested_options(
                options, ntf_fir_from_q0.default_options,
                [], ['scs_opts'], False)['scs_opts'])
        opts2['cvxpy_opts'].pop('override_kktsolver')
        dig_opts['cvxpy_opts'].update(opts2['cvxpy_opts'])
        from ._fir_weighting_cvxpy import (
            ntf_fir_from_digested as _ntf_fir_from_digested)
    elif opts2['modeler'] == 'cvxpy_old':
        dig_opts['cvxpy_tdr_opts'].update(digested_options(
            options, ntf_fir_from_q0.default_options,
            [], ['cvxopt_opts'], False)['cvxopt_opts'])
        from ._fir_weighting_cvxpy_tdr import (
            ntf_fir_from_digested as _ntf_fir_from_digested)
    elif opts2['modeler'] == 'picos':
        dig_opts['picos_opts'].update(digested_options(
            options, ntf_fir_from_q0.default_options,
            [], ['cvxopt_opts'], False)['cvxopt_opts'])
        from ._fir_weighting_picos import (
            ntf_fir_from_digested as _ntf_fir_from_digested)
    else:
        raise ValueError('Unsupported modeling backend {}'.format(
            opts2['modeler']))
    digested_options(options, {})
    # Do the computation
    poles = np.asarray(poles).reshape(-1)
    if len(poles) > 0:
        wn = mult_weightings(w, ([], poles, 1))
    else:
        wn = w
    q0 = q0_weighting(order, wn, **opts1)
    if normalize == 'auto':
        q0 = q0/q0[0]
    elif normalize is not None:
        q0 = q0*normalize
    if poles.shape[0] > order:
        raise ValueError('Too many poles provided')
    poles = padr(poles, order, 0)
    # Get denominator coefficients from a_1 to a_order (a_0 is 1)
    ar = np.poly(poles)[1:].real
    Q = la.toeplitz(q0)
    d, v = np.linalg.eigh(Q)
    if opts2['fix_pos']:
        d = d/np.max(d)
        d[d < 0] = 0.
    Qs = v.dot(np.diag(np.sqrt(d))).dot(v.T)
    A = np.eye(order, order, 1)
    A[order-1] = -ar[::-1]
    C = -ar[::-1].reshape((1, order))
    ntf_ir = _ntf_fir_from_digested(Qs, A, C, H_inf, **dig_opts)
    return (np.roots(ntf_ir), poles, 1.)

ntf_hybrid_weighting.default_options = {"modeler": "cvxpy_old",
                                        "cvxpy_opts": {'override_kktsolver':
                                                       True,
                                                       'solver': 'cvxopt'},
                                        "cvxopt_opts": {'maxiters': 100,
                                                        'abstol': 1e-7,
                                                        'reltol': 1e-6,
                                                        'feastol': 1e-6},
                                        'scs_opts': {'max_iters': 2500,
                                                     'eps': 1e-3,
                                                     'alpha': 1.8,
                                                     'normalize': True,
                                                     'use_indirect': False},
                                        'show_progress': True,
                                        'fix_pos': True}
ntf_hybrid_weighting.default_options.update(q0_weighting.default_options)


# Following part is deprecated

def q0_from_noise_weighting(P, w, **options):
    """
    Alias of :func:`q0_weighting`

    .. deprecated:: 0.11.0
        Function has been renamed :func:`q0_weighting`.
    """
    warn("Function superseded by q0_weighting in "
         "NTFdesign.weighting module", PyDsmDeprecationWarning)
    return q0_weighting(P, w, **options)


def synthesize_ntf_from_q0(q0, H_inf=1.5, normalize="auto", **options):
    """
    Alias of :func:`ntf_fir_from_q0`

    .. deprecated:: 0.11.0
        Function has been renamed :func:`ntf_fir_from_q0`.
    """
    warn("Function superseded by ntf_fir_from_q0 in "
         "NTFdesign.weighting module", PyDsmDeprecationWarning)
    return ntf_fir_from_q0(q0, H_inf, normalize, **options)


def synthesize_ntf_from_noise_weighting(order, noise_weighting, H_inf=1.5,
                                        normalize="auto", **options):
    """
    Alias of :func:`ntf_fir_weighting`

    .. deprecated:: 0.11.0
        Function has been renamed :func:`ntf_fir_weighting`.
    """
    warn("Function superseded by ntf_fir_weighting in "
         "NTFdesign module", PyDsmDeprecationWarning)
    return ntf_fir_weighting(order, noise_weighting, H_inf,
                             normalize, **options)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

from .sets import *
from .utils import *
from .functions import *
from .interface import *



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
Fourier transform related routines (:mod:`pydsm.ft`)
====================================================

Functions to compute the fft and the dtft

.. currentmodule:: pydsm.ft

Functions
---------

.. autosummary::
   :toctree: generated/

    fft_centered  -- FFT function where output vector has 0 frequency at center
    dtft  -- discrete time Fourier transform
    dtft_hermitian  -- DTFT specialized to hermitian vectors
    idtft -- inverse discrete time Fourier transform
    idtft_hermitian -- IDTFT specialized to hermitian vectors
"""

from __future__ import division, print_function

import numpy as np
import scipy as sp
__import__("scipy.fftpack")
__import__("scipy.integrate")
from .utilities import digested_options

__all__ = ["fft_centered", "dtft", "dtft_hermitian", "idtft",
           "idtft_hermitian"]


def fft_centered(x, fs=1):
    """
    Computes a *centered* FFT.

    Provides an FFT function where the output vector has the zero frequency
    at its center. This is more suitable for plotting.

    Parameters
    ----------
    x : array_like
        1-D sequence to compute the FFT upon

    Returns
    -------
    X : ndarray
        samples of the DFT of the input vector
    ff : ndarray
        vector of frequencies corresponding to the samples in X

    Other Parameters
    ----------------
    fs : real, optional
        sample frequency for the input vector (defaults to 1)
    """
    N = len(x)
    fs = float(fs)
    if np.mod(N, 2) == 0:
        k = np.arange(-N/2, N/2)  # N even
    else:
        k = np.arange(-(N-1)/2, (N-1)/2+1)  # N odd
    ff = k/(N/fs)
    X = sp.fftpack.fft(x)
    X = sp.fftpack.fftshift(X)
    return(ff, X)


def dtft(x, fs=1, t0=0):
    """
    Computes the discrete time Fourier transform (DTFT).

    Returns a function that is the DTFT of the given vector.

    Parameters
    ----------
    x :  array_like
        the 1-D vector to compute the DTFT upon

    Returns
    -------
    X : callable
        a function of frequency as in X(f), corresponding to the DTFT of x

    Other Parameters
    ----------------
    fs : real, optional
        sample frequency for the input vector (defaults to 1)
    t0 : real, optional
        the time when x[0] is sampled (defaults to 0). This is expressed
        in sample intervals.
    """
    return lambda f: np.sum(x*np.exp(-2j*np.pi*f/fs*(np.arange(len(x))-t0)))


def dtft_hermitian(x, fs=1):
    """
    Computes the discrete time Fourier transform of a hermitian vector.

    Since the input vector is hermitian, i.e. x[-n]=conjugate(x[n]), only
    a single side of it is sufficient for the computation of the DTFT,
    which is necessarily real.

    Parameters
    ----------
    x : array_like
        positive side of the 1-D sequence to compute the DTFT on.

    Returns
    -------
    X : callable
        a real function of frequency as in X(f), corresponding to the DTFT
        of x

    Other Parameters
    ----------------
    fs : real, optional
        sample frequency for the input vector (defaults to 1)
    """
    x = x.real
    return lambda f: \
        x[0]+np.sum(x[1:]*2*np.cos(2*np.pi*f/fs*np.arange(1, len(x))))


def _idtft(Ff, t, fs=1, **quad_opts):
    rr = sp.integrate.quad(lambda f: np.real(Ff(f*fs)*np.exp(2j*np.pi*f*t)),
                           -0.5, 0.5, **quad_opts)[0]
    ri = sp.integrate.quad(lambda f: np.imag(Ff(f*fs)*np.exp(2j*np.pi*f*t)),
                           -0.5, 0.5, **quad_opts)[0]
    return rr+1j*ri


def idtft(Ff, tt, fs=1, **options):
    """Compute the inverse discrete time Fourier transform (IDTFT)

    Parameters
    ----------
    Ff : callable
        the function of frequency to compute the IDTFT from
    tt : real or array_like
        the time sample (or 1-D vector of) where to compute the IDTFT

    Returns
    -------
    x : ndarray
        the inverse discrete time Fourier transform at time values in tt

    Other Parameters
    ----------------
    fs : real, optional
        the sample frequency for the output sequence (defaults to 1)
    quad_opts : dictionary
        Parameters to be passed to the ``quad`` function used internally as
        an integrator. Allowed options are ``epsabs``, ``epsrel``, ``limit``,
        ``points``. Do not use other options since they could break the
        integrator in unexpected ways. Defaults can be set by changing the
        function ``default_options`` attribute.

    See Also
    --------
    scipy.integrate.quad : integrator used internally.
        For the meaning of the integrator parameters.
    """
    # Manage optional parameters
    opts = digested_options(options, idtft.default_options, [], ['quad_opts'])
    # Do the computation
    if np.isscalar(tt):
        return _idtft(Ff, tt, fs, **opts['quad_opts'])
    else:
        return np.asarray([_idtft(Ff, t, fs, **opts['quad_opts']) for t in tt])

idtft.default_options = {"quad_opts": {"epsabs": 1E-12,
                                       "epsrel": 1E-9,
                                       "limit": 100,
                                       "points": None}}


def _idtft_hermitian(Ff, t, fs=1, **quad_opts):
    return 2*sp.integrate.quad(lambda f: np.real(Ff(f*fs))*np.cos(2*np.pi*f*t),
                               0, 0.5, **quad_opts)[0]


def idtft_hermitian(Ff, tt, fs=1, **options):
    """Compute the inverse discrete time Fourier transform (IDTFT) for a
    hermitian function of frequency.

    Since the input function is hermitian, i.e. F(-f)=conjugate(F(f)), it
    is sufficient to look at a single side of it. Furthermore, the IDTFT
    turns out to be a real vector.

    Parameters
    ----------
    Ff : callable
        the function of frequency to compute the IDTFT from. The values for
        a negative argument are not used.
    tt : real or array_like
        the time sample (or 1-D vector of) where to compute the IDTFT

    Returns
    -------
    x : ndarray
        the inverse discrete time Fourier transform at time values in tt

    Other Parameters
    ----------------
    fs : real, optional
        the sample frequency for the output sequence (defaults to 1)
    quad_opts : dictionary
        Parameters to be passed to the ``quad`` function used internally as
        an integrator. Allowed options are ``epsabs``, ``epsrel``, ``limit``,
        ``points``. Do not use other options since they could break the
        integrator in unexpected ways. Defaults can be set by changing the
        function ``default_options`` attribute.

    See Also
    --------
    scipy.integrate.quad : integrator used internally.
        For the meaning of the integrator parameters.
    """
    # Manage optional parameters
    opts = digested_options(options, idtft.default_options, [], ['quad_opts'])
    # Do the computation
    if np.isscalar(tt):
        return _idtft_hermitian(Ff, tt, fs, **opts['quad_opts'])
    else:
        return np.asarray([_idtft_hermitian(Ff, t, fs, **opts['quad_opts'])
                           for t in tt])

idtft_hermitian.default_options = idtft.default_options.copy()



#***********************************************************************#
# Copyright (C) 2010-2013 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
import cvxopt as opt
from ..defs import *
from ..scalars import cvxpy_obj
from ..scalars import cvxpy_scalar_var

# Class definition
class cvxpy_power_pos_epi(object):
    r"""
    | :math:`\{ (x,t) \in \mathbb{R}_+ \\times \mathbb{R} \ |
      \ x^p  \leq t \}` where :math:`p > 1`.
    """

    # Method: __init__
    def __init__(self,p):
        """
        Class constructor.
        """

        self.type = SET
        self.name = 'power_pos_epi'
        self.expansion_type = DIF
        self.p = p

    # Method: valid_shape
    def valid_shape(self,shape):

        return shape[1] == 1

    # Method: __str__
    def __str__(self):

        return self.name

    # Method: _construct
    def _construct(self,el,mp,n):

        p = self.p

        # Prepare zg
        if np.isscalar(el[0,0]):
            zg = el[0,0]
        elif type(el[0,0]) is cvxpy_obj:
            zg = el[0,0].value
        else:
            zg = None

        # Prepare tg
        if np.isscalar(el[1,0]):
            tg = el[1,0]
        elif type(el[1,0]) is cvxpy_obj:
            tg = el[1,0].value
        else:
            tg = None

        # f
        def f(x):
            z = zg
            t = tg
            if z is None:
                z = x[mp[el[0,0]]]
            if t is None:
                t = x[mp[el[1,0]]]
            return z**p-t

        # grad f
        def grad_f(x):
            z = zg
            g = opt.spmatrix(0.0,[],[],(n,1))
            if z is None:
                z = x[mp[el[0,0]]]
            if type(el[0,0]) is cvxpy_scalar_var:
                g[mp[el[0,0]]] = p*(z**(p-1.))
            if type(el[1,0]) is cvxpy_scalar_var:
                g[mp[el[1,0]]] = -1.
            return g

        # hess f
        def hess_f(x):
            z = zg
            h = opt.spmatrix(0.0,[],[],(n,n))
            if z is None:
                z = x[mp[el[0,0]]]
            if type(el[0,0]) is cvxpy_scalar_var:
                h[mp[el[0,0]],mp[el[0,0]]] = p*(p-1.)*(z**(p-2.))
            return h

        # ind f
        def ind_f(x):
            z = zg
            if z is None:
                z = x[mp[el[0,0]]]
            return z > 0.

        # Return functions
        return f, grad_f, hess_f, ind_f

# Create instance
power_pos_epi = cvxpy_power_pos_epi
